import asyncio
import time


class PaceMaker:
    def __init__(self, interval: float, init_burst: int = 0):
        self.interval = interval
        self.burst = init_burst
        self._last = 0

    def pace(self):
        if self.burst:
            self.burst -= 1
            return
        cur = time.time()
        delta = cur - self._last
        if delta < self.interval:
            time.sleep(self.interval - delta)
        self._last = max(cur, self._last + self.interval)

    async def apace(self):
        if self.burst:
            self.burst -= 1
            return
        cur = time.time()
        delta = cur - self._last
        if delta < self.interval:
            await asyncio.sleep(self.interval - delta)
        self._last = max(cur, self._last + self.interval)



import asyncio
import numpy as np
import platform
import soundfile
import tkinter as tk
from enum import Enum
from functools import partial
from sounddevice import OutputStream, CallbackStop
from threading import Lock, Thread
from tkinter.filedialog import askopenfile, asksaveasfile
from tkinter.messagebox import showerror
from tkinter.ttk import Scale, Separator, Combobox
from typing import MutableMapping, Callable, Optional

from ..common.config import default_config, config_path
from ..common.audio_devices import get_device_list, get_device_index
from ..common.constants import SAMPLING_RATE, FRAME_SIZE
from ..common.audio_files import load_audio_file, is_m4a_supported, load_m4a_file
from ..common.log import logger
from ..common.message_board import MessageBoard
from ..common.utils import my_assert, to_gain, remove_non_ascii, find_first
from .master_client import MasterClient, prepare_queue
from .mixing_desk import MixingDesk


class MasterClientState(Enum):
    IDLE = 0
    PREVIEW = 1
    PLAY = 2


class MasterClientGui:
    def __init__(self, config: MutableMapping[str, str], save_config: Optional[Callable[[], None]] = None):
        self.config = config
        self.save_config = save_config
        self.input = None
        self.output = None

        # GUI
        self.root = tk.Tk()
        self.root.title(f"Online Choir")
        self.open = tk.Button(self.root, text="Open", command=self._open_file)
        self.open.pack()
        self.start_label = tk.Label(text="0:00")
        self.start_label.pack()
        self.start_point = Scale(self.root, from_=0.0, to=1.0, state=tk.DISABLED, command=self._set_start_point)
        self.start_point.pack(fill=tk.X)
        self.preview = tk.Button(self.root, text="Play Locally", command=self._preview, state=tk.DISABLED)
        self.preview.pack()
        self.play = tk.Button(self.root, text="Play to Everyone", fg="red",
                              command=partial(self._start_session, live=False), state=tk.DISABLED)
        self.play.pack()
        self.live = tk.Button(self.root, text="Live Session", fg="red", command=partial(self._start_session, live=True))
        self.live.pack()
        self.stop = tk.Button(self.root, text="Stop", command=self._stop, state=tk.DISABLED)
        self.stop.pack()
        self.cur_time = tk.Label(text=" ")
        self.cur_time.pack()
        self.prev_time = None
        self.listen = tk.Button(self.root, text="Play Recording", command=self._listen, state=tk.DISABLED)
        self.listen.pack()
        self.save = tk.Button(self.root, text="Save Recording...", command=self._save, state=tk.DISABLED)
        self.save.pack()
        self.desk_button = tk.Button(self.root, text="Open Mixing Desk", command=self._open_desk, state=tk.DISABLED)
        self.desk_button.pack()
        self.desk = None

        # Device selection
        Separator(self.root).pack(fill=tk.X)
        device_sel = tk.Frame(self.root)
        tk.Label(device_sel, text="Input device:").grid(row=1, column=1)
        in_devs, in_idx = get_device_list(self.config, 'in_device', out=False)
        self.in_device_name = Combobox(device_sel, state='readonly', values=in_devs)
        self.in_device_name.current(in_idx)
        self.in_device_name.bind("<<ComboboxSelected>>", self._switch_input)
        self.in_device_name.grid(row=1, column=2)
        tk.Label(device_sel, text="Output device:").grid(row=2, column=1)
        out_devs, out_idx = get_device_list(self.config, 'device', out=True)
        self.out_device_name = Combobox(device_sel, state='readonly', values=out_devs)
        self.out_device_name.current(out_idx)
        self.out_device_name.bind("<<ComboboxSelected>>", self._switch_output)
        self.out_device_name.grid(row=2, column=2)
        device_sel.pack(pady=5, padx=10)

        # Server selection
        Separator(self.root).pack(fill=tk.X)
        server_sel = tk.Frame(self.root)
        tk.Label(server_sel, text="Server address:").grid(row=1, column=1)
        self.server_address = tk.Entry(server_sel, width=20)
        self.server_address.insert(0, self.config['server'])
        self.server_address.grid(row=1, column=2)
        tk.Label(server_sel, text="Server port:").grid(row=2, column=1)
        self.server_port = tk.Entry(server_sel, width=5)
        self.server_port.insert(0, self.config['port'])
        self.server_port.grid(row=2, column=2)
        server_sel.pack(pady=5)

        self.state = None

        # Audio
        self.stream = None
        self.play_mutex = Lock()
        self.play_cursor = None
        self.master_client = None

        # GUI updates
        self.update_board = MessageBoard()
        self.gui_updater = Thread(target=self._update_gui, name='GUI updater')
        self.gui_updater.start()

    def spin(self):
        self.state = MasterClientState.IDLE
        self.root.mainloop()
        if self.desk is not None:
            self.desk.changes.close()
        self.update_board.close()
        self.gui_updater.join()
        self.state = None  # nullify state after the GUI has stopped

    def _open_file(self):
        f_types = (
            ("WAV files", "*.wav"),
            ("Ogg files", "*.ogg"),
            ("FLAC", "*.flac"),
        )
        if platform.system() == "Darwin":
            f_types += (("Mp3 files", "*.mp3"),)
        if is_m4a_supported():
            f_types += (("M4a files", "*.m4a"),)
        logger.info("Selecting input file")
        f = askopenfile(mode="rb", parent=self.root, title="Open Audio File", filetypes=f_types)
        if f is not None:
            try:
                self._open_audio(f)
            except (ValueError, OSError, RuntimeError, soundfile.LibsndfileError) as e:
                logger.error(f"Failed to open file: {e}")
                showerror(title="Open file", message="Could not open file")
            finally:
                f.close()

    def _clear_stream(self):
        my_assert(self.state == MasterClientState.IDLE, "Should be idle")
        if self.stream is not None:
            my_assert(not self.stream.active, f"Stream active={self.stream.active}")
            self.stream.close()
            self.stream = None

    def _preview(self):
        self._clear_stream()
        s = float(self.start_point.get())
        logger.info(f"Preview audio file from {s:.1f}s")
        self.play_cursor = int(round(s * 44100 / FRAME_SIZE) * FRAME_SIZE)
        self._change_state(MasterClientState.PREVIEW)
        self.stream = OutputStream(device=self._get_out_device(),
                                   samplerate=SAMPLING_RATE,
                                   blocksize=FRAME_SIZE,
                                   channels=2,
                                   dtype="float32",
                                   callback=self._player_callback,
                                   finished_callback=self._player_done,
                                   latency='high')
        self.stream.start()

    def _listen(self):
        self._clear_stream()
        logger.info(f"Listen to recording")
        self.play_cursor = 0
        self._change_state(MasterClientState.PREVIEW)
        self.stream = OutputStream(device=self._get_out_device(),
                                   samplerate=SAMPLING_RATE,
                                   blocksize=FRAME_SIZE,
                                   channels=1,
                                   dtype="float32",
                                   callback=self._listen_callback,
                                   finished_callback=self._player_done,
                                   latency='high')
        self.stream.start()

    def _start_session(self, live: bool):
        self.config['server'], self.config['port'] = self.server_address.get(), self.server_port.get()
        self._clear_stream()
        level_db = float(self.config.get('monitor_level', -10))
        level_abs = to_gain(level_db) if level_db > -35 else 0
        logger.info(f"Starting session! (monitor level = {level_abs})")
        self.master_client = MasterClient(mute=live, monitor_level=level_abs, gui_board=self.update_board,
                                          in_device=self._get_in_device(), out_device=self._get_out_device())
        self._change_state(MasterClientState.PLAY)

        async def run_client():
            if live:
                q = None
            else:
                q = prepare_queue(self.input, float(self.start_point.get()))
            try:
                await self.master_client.main(self.config['server'], int(self.config['port']), q)
            except OSError as e:
                msg = f"OSError {find_first(e.args, lambda x: isinstance(x, str), 'unknown error')}"
                logger.error(msg)
                return msg
            else:
                logger.info("Master logic completed")
                return None

        def run_client_thread():
            msg = asyncio.run(run_client())
            logger.debug("Client thread completed")
            if msg is None:
                self.output = self.master_client.get_recording()
            else:
                showerror(title="Connection Error", message=msg)
            self.master_client = None
            self._change_state(MasterClientState.IDLE)

        Thread(target=run_client_thread, name='master client').start()

    def _stop(self):
        logger.info("Stopping")
        self.stop.config(state=tk.DISABLED)  # prevent re-clicking of the button while process is stopping.
        if self.state == MasterClientState.PREVIEW:
            with self.play_mutex:
                self.play_cursor = max(len(self.input) if self.input is not None else 0,
                                       len(self.output) if self.output is not None else 0)
        elif self.state == MasterClientState.PLAY:
            self.master_client.stop()

    def _save(self):
        logger.info("Saving output")
        f = asksaveasfile(mode="wb", parent=self.root, title="Save Recording As...", defaultextension=".wav")
        if f is not None:
            self._save_audio(f)
            f.close()

    def _change_state(self, new_state: MasterClientState):
        my_assert(new_state != self.state, "Repeated state")
        logger.debug(f"Changing state to {new_state}")

        # disabling buttons
        if new_state == MasterClientState.IDLE:
            self.stop.config(state=tk.DISABLED)
            if self.save_config is not None:
                self.save_config()
        else:  # MasterClientState.PLAY, MasterClientState.PREVIEW
            self.open.config(state=tk.DISABLED)
            self.play.config(state=tk.DISABLED)
            self.live.config(state=tk.DISABLED)
            self.preview.config(state=tk.DISABLED)
            self.listen.config(state=tk.DISABLED)
            self.save.config(state=tk.DISABLED)

        self.state = new_state  # change state

        # enabling buttons
        if new_state == MasterClientState.IDLE:
            self.open.config(state=tk.NORMAL)
            if self.input is not None:
                self.play.config(state=tk.NORMAL)
                self.preview.config(state=tk.NORMAL)
            self.live.config(state=tk.NORMAL)
            if self.output is not None:
                self.listen.config(state=tk.NORMAL)
                self.save.config(state=tk.NORMAL)
        else:  # MasterClientState.PLAY, MasterClientState.PREVIEW
            self.stop.config(state=tk.NORMAL)

    def _open_audio(self, fd):
        if is_m4a_supported() and fd.name.endswith('.m4a'):
            self.input = load_m4a_file(fd.name)
        else:
            self.input = load_audio_file(fd)
        self.root.title(fd.name.split('/')[-1])
        self.preview.config(state=tk.NORMAL)
        self.play.config(state=tk.NORMAL)
        self.start_point.config(to=len(self.input) / 44100, state=tk.NORMAL)
        self.start_point.set(0.0)

    def _save_audio(self, fd):
        soundfile.write(fd, self.output, SAMPLING_RATE, format="WAV")

    def _player_callback(self, outdata: np.ndarray, frames: int, _, status):
        if status:
            logger.error(f"Audio callback error {status}")
        with self.play_mutex:
            cur = self.play_cursor
            self.play_cursor = cur + FRAME_SIZE
        if cur >= len(self.input):
            outdata[:, :] = 0
            raise CallbackStop
        outdata[:, :] = self.input[cur: cur + FRAME_SIZE, :]
        self.update_board.post_message('tick', cur)

    def _listen_callback(self, outdata: np.ndarray, frames: int, _, status):
        if status:
            logger.error(f"Audio callback error {status}")
        with self.play_mutex:
            cur = self.play_cursor
            self.play_cursor = cur + FRAME_SIZE
        if cur >= len(self.output):
            outdata[:, :] = 0
            raise CallbackStop
        outdata[:, 0] = self.output[cur: cur + FRAME_SIZE]
        self.update_board.post_message('tick', cur)

    def _player_done(self):
        logger.debug(f"Player done, status is active:{self.stream.active}, stopped:{self.stream.stopped}")
        self._change_state(MasterClientState.IDLE)

    def _set_monitor_level(self, fader_value: str):
        level_db = float(fader_value) - 35
        self.config['monitor_level'] = str(level_db)
        if self.master_client is not None:
            level_abs = to_gain(level_db) if level_db > -35 else 0
            self.master_client.set_monitor_level(level_abs)

    def _set_start_point(self, p):
        s = float(p)
        self.start_label.config(text=f"{int(s) // 60}:{s % 60:02.1f}")

    def _set_cur_time(self, frame: int):
        tick = int(round(frame / 4410))
        if tick != self.prev_time:
            self.cur_time.config(text=f"{int(tick) // 600}:{(tick % 600) / 10:02.1f}")
            self.prev_time = tick

    def _switch_input(self, _):
        new_input = remove_non_ascii(self.in_device_name.get())
        logger.info(f"Changing audio input to {new_input}")
        self.config['in_device'] = new_input

    def _switch_output(self, _):
        new_output = remove_non_ascii(self.out_device_name.get())
        logger.info(f"Changing audio output to {new_output}")
        self.config['device'] = new_output

    def _get_in_device(self) -> int:
        return get_device_index(self.config.get('in_device', ''), False)

    def _get_out_device(self) -> int:
        return get_device_index(self.config.get('device', ''), True)

    def _open_desk(self):
        if self.desk is None:
            logger.debug("open mixing desk")
            self.desk = MixingDesk(self.root, self.config, self._close_desk, self._set_monitor_level)
        else:
            self.desk.top.lift()

    def _close_desk(self):
        if self.desk is None:
            return
        logger.debug("close mixing desk")
        self.desk.destroy()
        self.desk = None

    def _update_gui(self):
        logger.debug("start GUI updater")
        for topic, content in self.update_board.messages():
            if topic == 'tick':
                self._set_cur_time(content)
            elif topic == 'session_active':
                if content:
                    self.desk_button.config(state=tk.NORMAL)
                    if self.desk is not None:
                        self.desk.trigger_update()
                else:
                    self.desk_button.config(state=tk.DISABLED)
            # elif topic == 'levels':
            #     if self.desk:
            #         self.desk.set_channel_levels(content)
        logger.debug("GUI updater ended")


def main():
    logger.info(f"Started on {platform.system()}")
    config = default_config()
    if 'master' not in config.sections():
        config['master'] = {'server': 'online-choir.modetexte.ch',
                            'port': '8878'}

    def save_config():
        logger.debug("Saving config")
        with config_path().open('w') as fd:
            config.write(fd)

    gui = MasterClientGui(config['master'], save_config)
    logger.debug("Start master GUI")
    gui.spin()



import audioop
from pathlib import Path

import numpy as np
import threading
from queue import Queue, Empty
from subprocess import Popen, PIPE

from ..common.log import parse_args_and_set_logger, logger
from ..common.audio_codecs import StereoDecoder


NOISE_AMPLITUDE = 100
STEREO_NOISE = audioop.tostereo(np.random.randint(low=-NOISE_AMPLITUDE, high=NOISE_AMPLITUDE,
                                                  size=512, dtype='i2').tobytes(), 2, 1, 1)
ICES2_PATH = '/usr/bin/ices2'


class WebRadio:
    def __init__(self, config_file: str):
        self._q = Queue()
        self._ices = Popen([ICES2_PATH, config_file], stdin=PIPE)
        self._broadcast_thread = threading.Thread(target=self._broadcaster, name='radio broadcaster')
        self._broadcast_thread.start()
        self._decoder = StereoDecoder()

    @classmethod
    def is_available(cls) -> bool:
        return Path(ICES2_PATH).exists()

    def add_data(self, data: bytes):
        idx = StereoDecoder.peek_frame_index(data)
        if idx < 0:
            return
        if idx == 0:
            self._decoder = StereoDecoder()  # reset decoder
        left, right = self._decoder.decode_left_right(data)
        self._q.put_nowait(audioop.add(audioop.tostereo(left, 2, 1, 0), audioop.tostereo(right, 2, 0, 1), 2))

    def _broadcaster(self):
        while True:
            try:
                data = self._q.get_nowait()
                if data is None:
                    break
            except Empty:
                data = STEREO_NOISE
            self._ices.stdin.write(data)
        logger.debug("broadcaster completed")

    def close(self):
        self._q.put_nowait(None)
        self._broadcast_thread.join()
        self._ices.kill()


def cli():
    import wave
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('fname')
    parsed_args = parse_args_and_set_logger(parser)

    assert WebRadio.is_available()
    radio = WebRadio('/Users/olivier/tmp/ices-2.0.3/ices.xml')
    input("Press Enter to continue...")
    with wave.open(parsed_args.fname, 'rb') as wf:
        while True:
            data = wf.readframes(512)
            if not data:
                break
            radio.add_data(data)
    input("Press Enter to continue...")
    radio.close()
    exit(0)



import asyncio
import audioop
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Tuple, Optional, Iterable, Any

from ..common.log import logger
from ..common.platform import get_log_folder
from ..common.audio_codecs import AdpcmDecoder, AudioEncoder
from ..common.constants import CHUNKS_PER_SECOND, CHUNKS_IN_FLIGHT, SLAVE_PLAYBACK_BUFFER_SIZE, FrameType, FRAME_SIZE, \
    VOCAL_GROUPS
from ..common.utils import flush_queue, my_assert

MIX_WINDOW_SIZE = CHUNKS_IN_FLIGHT - SLAVE_PLAYBACK_BUFFER_SIZE - int(CHUNKS_PER_SECOND)
RAW_CHUNK_SIZE = 3 * FRAME_SIZE


@dataclass
class SourceInfo:
    name: str
    group: str
    gain: float
    decoder: Optional[AdpcmDecoder]
    buffer: Dict[int, bytes]

    @property
    def checked_in(self):
        return self.decoder is not None


class AudioMixer:
    def __init__(self, record_audio=False):
        self._sources: Dict[Tuple, SourceInfo] = {}
        self.mix_queue = asyncio.Queue()
        self.encoder = None
        self.mix_gain = None
        self.last_sources = None
        self.end_frame = None
        self.recording = [] if record_audio else None

    def register_source(self, source_id: Tuple, *, name: str = '', group: str = VOCAL_GROUPS[0], gain: float = 1.0):
        my_assert(source_id not in self._sources, f"source {source_id} already registered")
        self._sources[source_id] = SourceInfo(name, group, gain, None, {})

    def unregister_source(self, source_id: Tuple):
        was_active = self._sources[source_id].checked_in
        del self._sources[source_id]

        if self.mix_in_progress and was_active:  # if there are sill other sources, make sure the mix is up-to-date
            logger.debug(f"Trigger update after {source_id} left")
            self._update_mix()

    def reset(self):
        self.encoder = None
        self.mix_gain = None
        self.end_frame = None
        for source in self._sources.values():
            source.decoder = None
            source.buffer = {}
        flush_queue(self.mix_queue)
        if self.recording is not None:
            self.recording = []

    @property
    def num_sources(self):
        return len(self._sources)

    @property
    def num_active_sources(self):
        return sum(source.checked_in for source in self._sources.values())

    @property
    def mix_in_progress(self) -> bool:
        return self.encoder is not None

    def check_in(self, source_id: Tuple):
        logger.info(f"Source {source_id} checked in")
        self._sources[source_id].decoder = AdpcmDecoder()

    def check_out(self, source_id: Tuple):
        logger.info(f"Source {source_id} left the session (next_index was: "
                    f"{self._sources[source_id].decoder.next_index})")
        self._sources[source_id].decoder = None

    def _add_raw_frame(self, source_id: Tuple, idx: int, sig: bytes):
        if not self.mix_in_progress or idx >= self.encoder.next_index:  # store frames that are still useful
            buffer = self._sources[source_id].buffer
            my_assert(idx not in buffer, "duplicate frame idx")
            buffer[idx] = sig

    def inject_raw_frame(self, source_id: Tuple, idx: int, sig: bytes):
        my_assert(len(sig) == RAW_CHUNK_SIZE, "Wrong raw data size")
        decoder = self._sources[source_id].decoder
        my_assert(idx == decoder.next_index, "Wrong frame idx")
        decoder.next_index += 1  # fake decoding
        self._add_raw_frame(source_id, idx, sig)

    def add_frame(self, source_id: Tuple, frame: bytes):
        idx = AdpcmDecoder.peek_frame_index(frame)
        if idx == FrameType.KEEPALIVE:
            return   # ignore keepalive
        if not self._sources[source_id].checked_in:
            # accept joins while the mix has not been started
            if idx == FrameType.CHECK_IN:
                if not self.mix_in_progress:
                    self.check_in(source_id)
                else:
                    logger.warning(f"Source {source_id} checked in too late")
        else:
            if idx == FrameType.ABORT:  # drop from session
                self.check_out(source_id)
            else:
                exp = self._sources[source_id].decoder.decompress(frame, 3)
                self._add_raw_frame(source_id, idx, exp)

        # possibly produce a new mix frame
        self._update_mix()

    def set_end_frame(self, end_frame_idx: int):
        my_assert(self.end_frame is None, "duplicate end frame")
        logger.debug(f"Setting end frame to {end_frame_idx}")
        self.end_frame = end_frame_idx

    async def get_frame(self):
        frame = await self.mix_queue.get()
        return frame

    def _update_mix(self):
        """This function updates the status of the mix after a frame has been received or a client left"""
        if not self.mix_in_progress:
            # no mix in progress:
            # * check if all checked in
            # * if not, start if fastest client is about to dry out
            num_active_sources = self.num_active_sources
            if num_active_sources == self.num_sources:
                logger.debug("All in!")
                self._start_mix()
            elif num_active_sources > 0:
                bar = min(MIX_WINDOW_SIZE, self.end_frame or MIX_WINDOW_SIZE)
                _, idx_tip = self._quorum()
                if idx_tip >= bar:
                    self._start_mix()

        if self.mix_in_progress:
            # mix in progress:
            # * if all sources have dropped, close the mix
            # * see if mix frame(s) can be produced
            # * if all frames have been produced, close the mix
            # * drop used frames from buffers
            if self.num_active_sources == 0:
                # If all sources have left, interrupt
                logger.warning("All sources have left, interrupting")
                self.mix_queue.put_nowait(self.encoder.silence_frame(-2))
                return
            frame_ready = self._get_ready_idx()
            if frame_ready >= self.encoder.next_index:
                for idx in range(self.encoder.next_index, frame_ready + 1):
                    self._produce_frame(idx)
                if self.end_frame is not None and frame_ready == self.end_frame:
                    logger.info("Mixed last frame")
                    if self.recording:
                        self._dump_audio()

    def _start_mix(self):
        # * freeze mix gain
        my_assert(not self.mix_in_progress, "Mix already started")
        active_sources = {source_id: source.decoder.next_index
                          for source_id, source in self._sources.items() if source.checked_in}
        logger.info(f"Starting mix with {active_sources}")
        self.encoder = AudioEncoder()
        self.mix_gain = len(active_sources)
        self.last_sources = frozenset(active_sources)

    def _get_ready_idx(self):
        idx_ready, idx_tip = self._quorum()
        if self.end_frame is not None and idx_tip >= self.end_frame:
            # the fastest source has completed the track. Finish the mix.
            return self.end_frame
        return max(idx_ready, idx_tip - MIX_WINDOW_SIZE)

    def _produce_frame(self, idx: int):
        frames_to_mix = {}
        sources_status = {}
        for source_id, source in self._sources.items():
            if not source.checked_in:
                continue  # inactive source
            sources_status[source_id] = source.decoder.next_index
            if source.decoder.next_index <= idx:
                continue
            my_assert(idx in source.buffer, f"Frame {idx} not found in buffer of {source_id}")
            frames_to_mix[source_id] = (source.gain, source.buffer.pop(idx))
        active_sources = frozenset(frames_to_mix)
        if active_sources != self.last_sources:
            logger.debug(f"Producing mix frame {idx} with {len(frames_to_mix)} sources {sources_status}")
            self.last_sources = active_sources
        mix_frame = self.mix_frames(frames_to_mix.values(), self.mix_gain)
        self.mix_queue.put_nowait(self.encoder.create_frame(mix_frame))
        if self.recording is not None:
            self.recording.append(mix_frame)

    def _quorum(self) -> Tuple[int, int]:
        lasts = tuple(source.decoder.next_index - 1 for source in self._sources.values() if source.checked_in)
        return min(lasts), max(lasts)

    def _dump_audio(self):
        from common.audio_files import save_audio_file
        fname = get_log_folder() / f"recording_{datetime.now().strftime('%Y-%m-%d-%H:%M:%S')}.flac"
        logger.info(f"Saving mix under {fname}")
        save_audio_file(fname, b''.join(self.recording))
        self.recording = []

    @staticmethod
    def mix_frames(frames: Iterable[Tuple[float, bytes]], mix_gain: float) -> bytes:
        mix = None
        for gain, sig in frames:
            adj = audioop.mul(sig, 3, gain / mix_gain)
            if mix is None:
                mix = adj
            else:
                mix = audioop.add(mix, adj, 3)
        return audioop.lin2lin(mix, 3, 2)

    def process_command(self, command: Dict[str, Any]):
        action = command.get('action', 'none')
        if action == 'list':
            response = []
            for source_id, source in self._sources.items():
                if not source.checked_in:
                    continue
                response.append({'id': source_id, 'name': source.name, 'group': source.group,
                                 'gain': source.gain})
            logger.debug("sending list of active sources")
            return response
        if action == 'adjust_gain':
            source_id = tuple(command.get('id', ()))
            if source_id not in self._sources:
                logger.error(f"bad target {source_id}")
                return None
            try:
                gain = float(command['gain'])
            except (ValueError, KeyError):
                logger.error("invalid gain")
                return None
            logger.debug(f"change gain of source {source_id} to {gain}")
            self._sources[source_id].gain = gain
            return None
        logger.error(f"invalid action {action}")
        return None



import struct

from .constants import VOCAL_GROUPS
from .version import VERSION

HEADER_SIZE = 512
AUTH_KEY = b'\xb7#\xd2\x9b\xdf\x8b\xd0\xb4\x10\xdf\xdf\xfe^\xa6\xc7\x83W\xf1\xfe\xf8\xe7\xe1\xee\xac/\x90Un\x83\xe1\\"\x84\xe0\xaf\xa0:\x06\x90\xbe\xca\'\xe5\x80\x9c\x17\xfegcZ\t\x9e\x86lu\x11.?\xfcf\xa1\xbf\x9c\x1a\xef\x0c;\xb6r\x0e\xeaSk\x0e\x9c[\x81\x92X\xc2Uo\x82T\r\xc3\xba\xac@7\t\x8f\xca~\x8a}\xdd\rh\xa3\x05m\x7f\xbd\x83\xc4\x86c\xf0\\9\x15s\x1d\x8d\xf8\xf4\xf8\x81\xe4xU\xb6i\xe7w\xd9\xe4\xe8\xf6"\x10\xcd\x0e\xc7\xf9$\x84I;5\xb4K\x98\xae/d6\xce\x88?\x86\x94\\\xf1\xf9\xf8A\xa6\xd5p\x1c\xce>\xe9\xd9\x03\xbfu3D\x8f\xafTV6Zc\x85*\xb9\xec\xb8\xbb\xdc\x1fi\x89mb\xc5\x93-\xa7h \x99\xad\xfc\x86^\x9f\x94\xd2\xd3\xe5D\xa6\x0f\xcc\xb0\x10\x03\xe8-t\xed3\xc2\xc0\x85\x08\x03\x1ai\xdc]v\xacGG\xea\xccH\xfe\xc9UM&U\x0cj+!\x91a[I\xc8\x8c\xc9H\xd4M\xfaw'
PROTOCOL_VERSION = 3


class AuthenticationHeader:
    def __init__(self, name='', group='', client_version=VERSION, valid=True):
        self.is_valid = valid
        self.name = name
        self.group = group
        self.client_version = client_version

    @classmethod
    def bad_header(cls):
        return cls(valid=False)

    @classmethod
    def from_bytes(cls, data: bytes):
        version, = struct.unpack('>H', data[:2])
        key = data[HEADER_SIZE - len(AUTH_KEY):HEADER_SIZE]
        if key != AUTH_KEY:
            return cls.bad_header()
        if version == 1:
            name = ''
            group = VOCAL_GROUPS[0]
            client_version = 'Unknown (1)'
        elif version == 2:
            name = cls.decode_field(data[2:244], '')
            group = cls.decode_field(data[244:256], VOCAL_GROUPS[0])
            client_version = 'Unknown (2)'
        elif version == PROTOCOL_VERSION:
            name = cls.decode_field(data[2:220], '')
            client_version = cls.decode_field(data[220:244], 'Unknown (3)')
            group = cls.decode_field(data[244:256], VOCAL_GROUPS[0])
        else:
            return cls.bad_header()
        return cls(name, group, client_version)

    def to_bytes(self) -> bytes:
        version_b = struct.pack('>H', PROTOCOL_VERSION)
        name_b = self.encode_field(self.name, 218)
        client_version_b = self.encode_field(self.client_version, 24)
        group_b = self.encode_field(self.group, 12)
        return version_b + name_b + client_version_b + group_b + AUTH_KEY

    @staticmethod
    def encode_field(value: str, length: int) -> bytes:
        return value.encode('UTF8')[:length].ljust(length)

    @staticmethod
    def decode_field(data: bytes, default: str) -> str:
        try:
            return data.decode('UTF8').strip()
        except ValueError:
            return default



from threading import Event
from typing import Any, Tuple, Dict, Iterator


class MessageBoard:
    def __init__(self):
        self._event = Event()
        self._board = {}
        self._active = True

    def post_message(self, topic, content):
        self._board[topic] = content
        self._event.set()

    def post_messages(self, update: Dict):
        self._board.update(update)
        self._event.set()

    def close(self):
        self._active = False
        self._event.set()

    def _retrieve_message(self) -> Tuple[Any, Any]:
        topic, *_ = self._board
        content = self._board.pop(topic)
        return topic, content

    def get_message(self, timeout=None) -> Tuple[Any, Any]:
        while True:
            if self._board:
                return self._retrieve_message()
            ok = self._event.wait(timeout)
            if not ok:
                raise TimeoutError()
            self._event.clear()

    def messages(self, timeout=None) -> Iterator[Tuple[Any, Any]]:
        while self._active:
            while self._board:
                yield self._retrieve_message()
            ok = self._event.wait(timeout)
            if not ok:
                raise TimeoutError()
            self._event.clear()

    def flush_topic(self, topic):
        self._board.pop(topic, None)



# Code borrowed from sockio, https://github.com/tiagocoutinho/sockio
import asyncio
import socket

from .log import logger

IPTOS_NORMAL = 0x0
IPTOS_LOWDELAY = 0x10
IPTOS_THROUGHPUT = 0x08
IPTOS_RELIABILITY = 0x04
IPTOS_MINCOST = 0x02


def set_socket_settings(writer: asyncio.StreamWriter, no_delay=True, tos=IPTOS_LOWDELAY, keep_alive=None):
    sock = writer.transport.get_extra_info("socket")
    try:
        if hasattr(socket, "TCP_NODELAY") and no_delay:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "IP_TOS"):
            sock.setsockopt(socket.SOL_IP, socket.IP_TOS, tos)
        if keep_alive is not None and hasattr(socket, "SO_KEEPALIVE"):
            if isinstance(keep_alive, (int, bool)):
                keep_alive = 1 if keep_alive in {1, True} else False
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, keep_alive)
            else:
                active = keep_alive.get('active')
                idle = keep_alive.get('idle')  # aka keepalive_time
                interval = keep_alive.get('interval')  # aka keepalive_intvl
                retry = keep_alive.get('retry')  # aka keepalive_probes
                if active is not None:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, active)
                if idle is not None:
                    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, idle)
                if interval is not None:
                    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, interval)
                if retry is not None:
                    sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, retry)
    except OSError as e:
        logger.warning(f"Failed to set the socket options ({e})")



from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Optional
import tkinter as tk

# TODO: create abstract class and DotLevelMeter

from .signal_smoother import SignalSmoother


class LevelMeter(ABC):
    """Abstract class for level meters."""

    @abstractmethod
    def update_level(self, db: Optional[float]) -> None:
        """By convention, None is used to indicate no signal."""
        pass


class DotLevelMeter(LevelMeter):
    """Level meter that displays a colored dot."""

    class SignalLevel(IntEnum):
        white = -1
        grey = 0
        green = 1
        orange = 2
        red = 3

    def __init__(self, master, **kwargs):
        self._canvas = tk.Canvas(master=master, width=20, height=20)
        self._dot = self._canvas.create_oval(10, 10, 20, 20, fill='white')
        self._smoother = SignalSmoother(self.SignalLevel.white)
        self._canvas.grid(**kwargs)

    def _db_to_color(self, db: float) -> SignalLevel:
        if db < -10:
            return self.SignalLevel.grey
        if db < -4:
            return self.SignalLevel.green
        if db <= 0:
            return self.SignalLevel.orange
        return self.SignalLevel.red

    def update_level(self, db: Optional[float]) -> None:
        if db is None:
            self._smoother.reset()
            self._set_color(self.SignalLevel.white)
            return
        color = self._db_to_color(db)
        color = self._smoother(color)
        if color is None:
            return
        self._set_color(color)

    def _set_color(self, color: SignalLevel) -> None:
        self._canvas.itemconfigure(self._dot, fill=color.name)


class BarLevelMeter(LevelMeter):
    """
    A dB meter that shows the signal level as an array of 10 dots.

       grey                |  green  | orange  | red
    O....O....O....O....O....O....O....O....O....O
    -inf -21  -18  -15  -12  -9   -6   -3   0   saturation
    """
    class LED:
        def __init__(self, canvas, idx: int):
            self.canvas = canvas
            self.tag = canvas.create_oval(10 + 20 * idx, 10, 20 + 20 * idx, 20, fill='white')
            if idx < 5:
                self.color = 'grey'
            elif idx < 7:
                self.color = 'green'
            elif idx < 9:
                self.color = 'orange'
            else:
                self.color = 'red'

        def on(self):
            self.canvas.itemconfigure(self.tag, fill=self.color)

        def off(self):
            self.canvas.itemconfigure(self.tag, fill='white')

    def __init__(self, master, **kwargs):
        canvas = tk.Canvas(master=master, width=210, height=20)
        self.dots = [self.LED(canvas, i) for i in range(10)]
        self.prev_level = 0
        self._smoother = SignalSmoother(0)
        canvas.grid(**kwargs)

    @staticmethod
    def _db_to_level(db: float) -> int:
        """Convert a dB value to a level index between 1 and 10 (included)."""
        if db > 0:
            return 10
        # 0 and 10 are reserved values, hence the calculation may only return values between 1 and 9
        return max(1, min(9, int(round(db / 3 + 9))))

    def update_level(self, db: Optional[float]):
        if db is None:
            self.reset()
        level = self._smoother(self._db_to_level(db))
        if level is None:  # no change needed
            return
        if level > self.prev_level:
            for dot in range(self.prev_level, level):
                self.dots[dot].on()
        if level < self.prev_level:
            for dot in range(level, self.prev_level):
                self.dots[dot].off()
        self.prev_level = level

    def reset(self):
        for dot in range(0, self.prev_level):
            self.dots[dot].off()
        self.prev_level = 0
        self._smoother.reset()



import gzip
import json
import struct


class MessageProtocol:
    def __init__(self, msg):
        self.msg = msg

    @staticmethod
    def get_length(data: bytes) -> int:
        assert len(data) == 2
        length, = struct.unpack('>H', data)
        return length

    @classmethod
    def from_bytes(cls, data: bytes):
        return cls(json.loads(gzip.decompress(data).decode('UTF8')))

    def to_bytes(self) -> bytes:
        payload = gzip.compress(json.dumps(self.msg).encode('UTF8'))
        len_b = struct.pack('>H', len(payload))
        return len_b + payload



"""
The server backend.
"""
import asyncio
import audioop
from collections import deque
from typing import Optional

from ..common.audio_codecs import StereoEncoder, AudioEncoder, StereoDecoder, AdpcmDecoder
from ..common.constants import CONNECTION_TIMEOUT, SLAVE_PLAYBACK_BUFFER_SIZE, CHUNKS_PER_SECOND
from ..common.version import VERSION
from ..common.auth import HEADER_SIZE, AuthenticationHeader
from ..common.log import logger
from ..common.message_protocol import MessageProtocol
from ..common.network import set_socket_settings
from ..common.utils import iterate_stream, my_assert
from .audio_broadcaster import AudioBroadcaster
from .audio_mixer import AudioMixer
from .web_radio import WebRadio

CHECK_IN_TIMEOUT = 5  # seconds
MONITOR_QUEUE_SIZE = SLAVE_PLAYBACK_BUFFER_SIZE + int(CHUNKS_PER_SECOND) + 1


class MixServer:
    def __init__(self, master_port: int, slave_port: int, control_port: int, *,
                 ices_config: Optional[str] = None,
                 monitor_level: Optional[float] = None,
                 record_audio: bool = False,
                 safe: bool = True):
        self.mixer = AudioMixer(record_audio)
        self.broadcaster = AudioBroadcaster()
        self.radio = None
        self.monitor_level = monitor_level
        self.ready_for_session = asyncio.Event()
        self._master_port = master_port
        self._slave_port = slave_port
        self._control_port = control_port
        self._ices_config = ices_config
        self._master_server = None
        self._slave_server = None
        self._control_server = None
        self._safe = safe

    async def _send_mix_to_master(self, writer):
        try:
            while True:
                data = await self.mixer.get_frame()
                writer.write(data)
                await writer.drain()
        except ConnectionResetError:
            logger.debug("connection to master reset")

    async def _forward_audio_to_client(self, writer, source_id):
        while True:
            data = await self.broadcaster.get_frame(source_id)
            writer.write(data)
            await writer.drain()

    async def _serve_master(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        set_socket_settings(writer)
        try:
            header = AuthenticationHeader.from_bytes(await reader.readexactly(HEADER_SIZE))
        except (asyncio.IncompleteReadError, ConnectionResetError):
            logger.warning("Master did not send auth packet")
            writer.close()
            return
        if not header.is_valid:
            logger.warning("Invalid master client connected")
            writer.close()
            return
        if not self.ready_for_session.is_set():
            logger.warning("Second master attempted to join")
            writer.write(AudioEncoder.silence_frame(-2))  # terminate session immediately
            await writer.drain()
            return
        if self.mixer.num_sources == 0:
            logger.warning("Master attempted to join while there is no source")
            writer.write(AudioEncoder.silence_frame(-2))  # terminate session immediately
            await writer.drain()
            return
        logger.info(f"Master ({header.client_version}) joined from {writer.get_extra_info('peername')}")
        self.ready_for_session.clear()
        self.mixer.reset()
        self.broadcaster.reset()
        if self.monitor_level is not None:
            self.mixer.register_source(("monitor",), name='Monitor', gain=self.monitor_level)
            self.mixer.check_in(("monitor",))  # check in immediately
            decoder = StereoDecoder()
            monitor_queue = deque(maxlen=MONITOR_QUEUE_SIZE)
        mix_sender = asyncio.create_task(self._send_mix_to_master(writer))

        def check_for_active_slaves():
            logger.debug("checking for active sources")
            if self.mixer.num_active_sources == 0:
                logger.warning("No slave checked in, abort")
                mix_sender.cancel()
                writer.write(AudioEncoder.silence_frame(-2))  # terminate session immediately
                writer.close()
                self.broadcaster.reset()

        last_frame = -1
        got_end_frame = False
        checker = None
        async for data in iterate_stream(reader, StereoDecoder.PACKET_SIZE):
            idx = StereoDecoder.peek_frame_index(data)
            if idx < 0:
                logger.debug("  got end frame")
                got_end_frame = True
                self.mixer.set_end_frame(last_frame)
            else:
                my_assert(idx == last_frame + 1, "wrong idx")
                if idx == 0:
                    # Launch timer to check
                    logger.debug("Starting timer for check-ins")
                    checker = asyncio.get_event_loop().call_later(CHECK_IN_TIMEOUT, check_for_active_slaves)
                last_frame = idx
                if self.monitor_level is not None:  # send audio to the mix, too
                    left, right = decoder.decode_left_right(data, 3)
                    sig = audioop.add(audioop.mul(left, 3, 0.5), audioop.mul(right, 3, 0.5), 3)
                    if len(monitor_queue) == monitor_queue.maxlen:
                        self.mixer.inject_raw_frame(("monitor",), *monitor_queue.pop())
                    monitor_queue.appendleft((idx, sig))
            self.broadcaster.add_input_frame(data)
            if self.radio:
                self.radio.add_data(data)

        # connection closed
        if not got_end_frame:  # this only happens when the master drops abruptly
            logger.warning("  master closed early")
            self.broadcaster.add_input_frame(StereoEncoder.silence_frame(-1))
            self.mixer.set_end_frame(last_frame)
        mix_sender.cancel()  # Normally master closes when the whole mix has been received
        if self.monitor_level is not None:
            self.mixer.unregister_source(("monitor",))  # monitor the audio sent
        self.ready_for_session.set()
        if checker is not None:
            checker.cancel()
        logger.info("Master left")

    async def _serve_slave(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        set_socket_settings(writer)
        try:
            header = AuthenticationHeader.from_bytes(await reader.readexactly(HEADER_SIZE))
        except (asyncio.IncompleteReadError, ConnectionResetError):
            logger.warning("Slave did not send auth packet")
            writer.close()
            return
        if not header.is_valid:
            logger.warning("Invalid slave client connected")
            writer.close()
            return
        client_id = writer.get_extra_info('peername')
        logger.info(f"Client {client_id} ({header.name}, {header.group}, {header.client_version}) connected")
        self.broadcaster.register_client(client_id)
        self.mixer.register_source(client_id, name=header.name, group=header.group)
        audio_sender = asyncio.create_task(self._forward_audio_to_client(writer, client_id))
        try:
            async for data in iterate_stream(reader, AdpcmDecoder.PACKET_SIZE, timeout=CONNECTION_TIMEOUT):
                self.mixer.add_frame(client_id, data)
        except asyncio.CancelledError:
            logger.warning(f"Closing connection with {client_id}")
        except Exception as e:
            logger.info(f"Connection with {client_id} lost (error: {e})")
        finally:
            audio_sender.cancel()
            self.mixer.unregister_source(client_id)
            self.broadcaster.unregister_client(client_id)
            logger.info(f"Client {client_id} disconnected")
            writer.close()

    async def _serve_controller(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        set_socket_settings(writer)
        try:
            header = AuthenticationHeader.from_bytes(await reader.readexactly(HEADER_SIZE))
        except (asyncio.IncompleteReadError, ConnectionResetError):
            logger.warning("Controller did not send auth packet")
            writer.close()
            return
        if not header.is_valid:
            logger.warning("Invalid controller connected")
            writer.close()
            return
        logger.info(f"Controller ({header.client_version}) joined from {writer.get_extra_info('peername')}")
        while True:
            try:
                l = await reader.readexactly(2)
                msg = MessageProtocol.from_bytes(await reader.readexactly(MessageProtocol.get_length(l)))
                logger.debug(f"Received command {msg.msg}")
            except (ConnectionResetError, ValueError, asyncio.IncompleteReadError):
                break
            response = self.mixer.process_command(msg.msg)
            if response is not None:
                writer.write(MessageProtocol(response).to_bytes())
                await writer.drain()
        logger.info("Controller left")

    async def __aenter__(self):
        self._master_server = await asyncio.start_server(self._serve_master, port=self._master_port)
        self._slave_server = await asyncio.start_server(self._serve_slave, port=self._slave_port)
        self._control_server = await asyncio.start_server(self._serve_controller, port=self._control_port)

        if self._ices_config is not None and WebRadio.is_available():
            self.radio = WebRadio(self._ices_config)
        await self._master_server.__aenter__()
        await self._slave_server.__aenter__()
        await self._control_server.__aenter__()
        logger.info(f"Server v{VERSION} started")
        self.ready_for_session.set()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logger.info("Stopping server")

        # wait for session to complete
        if self._safe and not self.ready_for_session.is_set():
            logger.info("Session in progress, waiting...")
            await self.ready_for_session.wait()

        # close radio channel
        if self.radio:
            self.radio.close()

        # ramp down servers
        await self._control_server.__aexit__(exc_type, exc_val, exc_tb)
        await self._slave_server.__aexit__(exc_type, exc_val, exc_tb)
        await self._master_server.__aexit__(exc_type, exc_val, exc_tb)

    async def serve_forever(self):
        await self._master_server.serve_forever()
        await self._slave_server.serve_forever()
        await self._control_server.serve_forever()


async def start_server(monitor_level: Optional[float],
                       master_port: int,
                       slave_port: int,
                       control_port: int,
                       record_audio: bool,
                       ices_config: Optional[str]):
    server_object = MixServer(master_port, slave_port, control_port,
                              record_audio=record_audio,
                              monitor_level=monitor_level,
                              ices_config=ices_config)
    async with server_object:
        await server_object.serve_forever()



from collections import deque

DEFAULT_STICKINESS = 20


class SignalSmoother:
    def __init__(self, initial_state: int, stickiness=DEFAULT_STICKINESS):
        self._prev_inputs = deque(maxlen=stickiness + 1)
        self._prev_output = initial_state
        self._initial_state = initial_state

    def __call__(self, level: int):
        self._prev_inputs.append(level)
        ret_value = max(self._prev_inputs)
        if ret_value != self._prev_output:
            self._prev_output = ret_value
            return ret_value
        return None

    def reset(self):
        self._prev_inputs = deque(maxlen=self._prev_inputs.maxlen)
        self._prev_output = self._initial_state




def app():
    """Entry point for the packaged (pyinstaller) app"""
    from onlinechoir.common.log import setup_logging
    from onlinechoir.common.platform import get_log_folder
    from .gui import main
    setup_logging('ERROR', get_log_folder() / "master_gui.log")
    main()


def cli():
    """Entry point for the pure python master client"""
    import argparse

    from onlinechoir.common.log import parse_args_and_set_logger
    from .master_client import main as main_cli
    from .gui import main as main_gui

    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--no-gui', action="store_true",
                        help="Run the command-line client")
    parser.add_argument('-i', '--input',
                        help="(CLI only) The input audio file")
    parser.add_argument('-s', '--server', default='online-choir.modetexte.ch',
                        help="(CLI only) The address of the server")
    parser.add_argument('-p', '--port', type=int, default=8878,
                        help="(CLI only) The port to connect to (master port)")
    parser.add_argument('-o', '--output',
                        help="(CLI only) Record the session into the given file (WAV or FLAC)")
    parser.add_argument('-x', '--mute', action="store_true",
                        help="(CLI only) Do not play any sound")
    parser.add_argument('-l', '--live', action='store_true',
                        help="(CLI only) Start a live session: use audio input instead of a file")
    parser.add_argument('-m', '--monitor-level', type=float, default=0.1,
                        help="(CLI only) Mix the audio file with the choir with the given gain")
    parser.add_argument('-k', '--skip-seconds', type=float,
                        help="(CLI only) Skip this number of seconds of the audio file")

    parsed_args = parse_args_and_set_logger(parser, 'master')

    if parsed_args.no_gui:
        main_cli(inpt=parsed_args.input,
                 server=parsed_args.server,
                 port=parsed_args.port,
                 output=parsed_args.output,
                 mute=parsed_args.mute,
                 live=parsed_args.live,
                 monitor_level=parsed_args.monitor_level,
                 skip_seconds=parsed_args.skip_seconds)
    else:
        main_gui()



import asyncio
from typing import Dict

from ..common.audio_codecs import StereoDecoder, StereoEncoder
from ..common.constants import KEEPALIVE_INTERVAL, FrameType
from ..common.utils import my_assert


class AudioBroadcaster:
    def __init__(self):
        self.clients: Dict[str, asyncio.Queue] = {}
        self.last_sent = -1

    def register_client(self, client_id: str) -> int:
        my_assert(client_id not in self.clients, f"client {client_id} already registered")
        self.clients[client_id] = asyncio.Queue()
        return self.last_sent

    def unregister_client(self, client_id: str):
        del self.clients[client_id]

    def reset(self):
        for queue in self.clients.values():
            while not queue.empty():
                queue.get_nowait()
        self.last_sent = -1

    def add_input_frame(self, frame: bytes):
        for client_id, queue in self.clients.items():
            queue.put_nowait(frame)
        idx = StereoDecoder.peek_frame_index(frame)
        my_assert(idx == -1 or idx == self.last_sent + 1, "unexpected frame idx")
        self.last_sent = idx

    async def get_frame(self, client_id: str):
        """Gets the next frame to send. This is possibly a keepalive frame."""
        queue = self.clients[client_id]
        try:
            frame = await asyncio.wait_for(queue.get(), timeout=KEEPALIVE_INTERVAL)
        except asyncio.TimeoutError:
            frame = StereoEncoder.silence_frame(FrameType.KEEPALIVE)
        return frame



from configparser import ConfigParser
from pathlib import Path

from .platform import get_config_folder


def read_config(fname: Path):
    conf = ConfigParser()
    conf.read(str(fname))
    return conf


def config_path() -> Path:
    return get_config_folder() / 'config.ini'


def default_config():
    return read_config(config_path())



"""
The client for the choir director.
"""
import asyncio
import numpy as np
import queue
try:
    import sounddevice
except (OSError, ModuleNotFoundError):
    pass
from functools import partial
from threading import Event, Lock
from typing import Optional, List

from ..common.utils import my_assert, iterate_queue
from ..common.log import logger
from ..common.audio_codecs import AudioDecoder, StereoEncoder
from ..common.constants import SAMPLING_RATE, FRAME_SIZE, CHUNKS_IN_FLIGHT, MASTER_PLAYBACK_BUFFER_SIZE, FrameType
from ..common.version import VERSION
from ..common.auth import AuthenticationHeader
from ..common.message_board import MessageBoard
from ..common.network import set_socket_settings


def prepare_queue(sig: np.ndarray, skip_seconds: Optional[float] = None) -> asyncio.Queue:
    n_frames, n_channels = sig.shape
    my_assert(n_channels == 2, "Stereo signal required")
    my_assert(n_frames % FRAME_SIZE == 0, "Signal must be adjusted to the chunk size")
    start_frame = int(round(skip_seconds * 44100 / FRAME_SIZE) * FRAME_SIZE) if skip_seconds else 0
    q = asyncio.Queue()
    for i in range(start_frame, n_frames, FRAME_SIZE):
        q.put_nowait(sig[i:i + FRAME_SIZE, :])
    q.put_nowait(None)
    return q


class MasterClient:
    def __init__(self, *,
                 mute: bool,
                 monitor_level: float,
                 gui_board: Optional[MessageBoard] = None,
                 in_device: Optional[int] = None, out_device: Optional[int] = None):
        logger.debug(f"Init master client with mute = {mute}, monitor_level = {monitor_level}, "
                     f"in_device = {in_device}, out_device={out_device}")
        my_assert(0 <= monitor_level <= 1, "Monitor level should be in [0, 1]")
        self.loop = None
        self.in_device = in_device
        self.out_device = out_device
        self.audio_in: List[np.ndarray] = []
        self.mute = mute
        self.monitor_mutex = Lock()
        self.monitor_level = monitor_level
        self.gui_board = gui_board

        # Playing status
        self.new_frame_received = None  # to be filled from main()
        self.encoder = StereoEncoder()
        self.stop_flag = False
        self.audio_mix: Optional[asyncio.Queue] = None  # to be filled from main()

    async def send_file(self, tcp_conn: asyncio.StreamWriter, data: asyncio.Queue):
        async for chunk in iterate_queue(data):
            while len(self.audio_in) < self.encoder.next_index - CHUNKS_IN_FLIGHT and not self.stop_flag:
                await self.new_frame_received.wait()
                self.new_frame_received.clear()
            if self.stop_flag:
                break
            tcp_conn.write(self.encoder(chunk))
            self.audio_mix.put_nowait(chunk.mean(axis=1))
            await tcp_conn.drain()

        tcp_conn.write(StereoEncoder.silence_frame(FrameType.EOF))
        await tcp_conn.drain()
        logger.info("Done sending file")

    async def play_stream(self, tcp_conn: asyncio.StreamReader):
        player_queue = queue.Queue()
        loop = asyncio.get_event_loop()
        playback_ended = asyncio.Event()
        decoder = AudioDecoder()
        paused = Event()
        paused.set()  # start the stream in pause mode
        playback_ended.set()  # set playback_ended, in case the session is aborted and the player never starts

        def player_callback(outdata: np.ndarray, frames: int, _, status):
            assert frames == FRAME_SIZE
            if status:
                logger.error(f"Audio callback error {status}")
            if paused.is_set():  # if not playing, play silence
                outdata[:, 0] = 0
            else:
                try:
                    chunk = player_queue.get_nowait()
                except queue.Empty:
                    logger.warning("Queue ran empty. Re-buffering")
                    paused.set()
                    outdata[:, 0] = 0
                    return
                if chunk is None:
                    raise sounddevice.CallbackStop
                with self.monitor_mutex:
                    monitor_level = self.monitor_level
                mon_chunk = self.audio_mix.get_nowait()
                outdata[:, 0] = self.audio_in[chunk] + monitor_level * mon_chunk
                self._update_gui('tick', chunk * FRAME_SIZE)

        def finished_callback():
            if loop.is_running():
                loop.call_soon_threadsafe(playback_ended.set)

        if not self.mute:
            stream = sounddevice.OutputStream(samplerate=SAMPLING_RATE, blocksize=FRAME_SIZE, channels=1,
                                              dtype="float32", callback=player_callback,
                                              finished_callback=finished_callback,
                                              device=self.out_device)
            stream.start()
        while len(self.audio_in) < max(1, self.encoder.next_index):
            try:
                data = await tcp_conn.readexactly(decoder.PACKET_SIZE)
            except asyncio.IncompleteReadError:
                data = None
            if not data:
                logger.info("Connection closed")
                break
            idx = decoder.peek_frame_index(data)
            if idx <= 0:
                logger.debug(f"Got frame {idx}")
                if idx == 0:
                    self._update_gui('session_active', True)
            if idx == -2:  # early drop message
                logger.debug("got early drop")
                break
            chunk = np.empty((FRAME_SIZE,))
            decoder(data, chunk)
            player_queue.put(len(self.audio_in))
            self.audio_in.append(chunk)
            self.new_frame_received.set()
            if not self.mute and paused.is_set() and player_queue.qsize() >= MASTER_PLAYBACK_BUFFER_SIZE:
                logger.info("Start playing after buffer reached sufficient level")
                playback_ended.clear()
                paused.clear()
        logger.info(f"Received {len(self.audio_in)} chunks (encoder at {self.encoder.next_index})")
        self._update_gui('session_active', False)
        if not self.mute:
            if paused.is_set() and not player_queue.empty():
                logger.info("Start playing after all chunks are received")
                playback_ended.clear()
                paused.clear()
            player_queue.put(None)
            logger.info("Waiting for audio to finish playing")
            await playback_ended.wait()
            stream.close()

    async def main(self, server_address: str, server_port: int, signal: Optional[asyncio.Queue]):
        msg = "live" if signal is None else f"{signal.qsize()} chunks"
        logger.info(f"Start client v{VERSION}, playing {msg}")
        self.loop = asyncio.get_event_loop()
        self.new_frame_received = asyncio.Event()
        logger.debug(f"Connecting to {server_address} port {server_port}")
        reader, writer = await asyncio.open_connection(server_address, server_port)
        set_socket_settings(writer)
        writer.write(AuthenticationHeader().to_bytes())
        await writer.drain()
        if signal is None:
            n_chan = min(2, sounddevice.query_devices(self.in_device, kind='input')['max_input_channels'])
            self.mute = True  # force mute mode to avoid audio
            signal = asyncio.Queue()
            stream_ended = asyncio.Event()
            frame_count = 0

            def rec_callback(indata: np.ndarray, frames: int, _, status):
                nonlocal frame_count
                chunk = np.empty((FRAME_SIZE, 2))
                sig = indata.mean(axis=1)
                chunk[:, 0] = sig
                chunk[:, 1] = sig
                self.loop.call_soon_threadsafe(partial(signal.put_nowait, chunk))
                self._update_gui('tick', frame_count)
                frame_count += frames

            def end_callback():
                self.loop.call_soon_threadsafe(stream_ended.set)

            stream = sounddevice.InputStream(SAMPLING_RATE, FRAME_SIZE, channels=n_chan, dtype='float32',
                                             latency='high', device=self.in_device, callback=rec_callback,
                                             finished_callback=end_callback)
            stream.start()
        else:
            stream = None
        self.audio_mix = asyncio.Queue()
        sender_task = asyncio.create_task(self.send_file(writer, signal))
        await self.play_stream(reader)
        if stream:
            stream.stop()
            await stream_ended.wait()
            stream.close()
        sender_task.cancel()
        writer.close()

    def _stop(self):
        """Stop the session immediately and return"""
        logger.info("Stopping the session")
        self.stop_flag = True
        self.new_frame_received.set()

    def stop(self):
        self.loop.call_soon_threadsafe(self._stop)

    def set_monitor_level(self, level: float):
        my_assert(0 <= level <= 1, f"Invalid level {level}")
        with self.monitor_mutex:
            self.monitor_level = level

    def get_recording(self) -> Optional[np.ndarray]:
        if not self.audio_in:
            return None
        res = np.empty((len(self.audio_in) * FRAME_SIZE,))
        for i, chunk in enumerate(self.audio_in):
            res[i * FRAME_SIZE:(i + 1) * FRAME_SIZE] = chunk
        return res

    def _update_gui(self, topic: str, msg):
        if self.gui_board is not None:
            self.gui_board.post_message(topic, msg)


def main(*, inpt: Optional[str], server: str, port: int, output: Optional[str], mute: bool, live: bool,
         monitor_level: float, skip_seconds: Optional[float]):
    import signal

    if inpt is None and not live:
        print("Must either provide an input file or choose the live option")
        exit(1)

    client = MasterClient(mute=mute, monitor_level=monitor_level)

    async def run_client():
        loop = asyncio.get_event_loop()
        try:
            loop.add_signal_handler(signal.SIGINT, client._stop)
        except NotImplementedError:
            pass
        if live:
            q = None
        else:
            from onlinechoir.common.audio_files import load_audio_file
            q = prepare_queue(load_audio_file(inpt), skip_seconds)
        await client.main(server, port, q)
        return

    asyncio.run(run_client())

    if output is not None:
        res = client.get_recording()
        if res is not None:
            import soundfile
            soundfile.write(output, res, SAMPLING_RATE)
    exit(0)



import os
import psutil
from pathlib import Path

from .log import logger
from .platform import get_log_folder


def _pid_file_path(name: str) -> Path:
    return get_log_folder() / (name + '.pid')


def _check_other_processes(name: str, pid: int, process_name: str):
    pid_file = _pid_file_path(name)
    if not pid_file.exists():
        return
    tokens = pid_file.read_text().split(",")
    if len(tokens) != 2:
        return
    try:
        prev_pid = int(tokens[0])
    except ValueError:
        return  # ignore malformed files
    prev_process_name = tokens[1]
    try:
        prev_process = psutil.Process(prev_pid)
    except psutil.NoSuchProcess:
        return
    if prev_pid == os.getpid():
        return  # either own process or dead process
    if prev_process.name() != prev_process_name:
        return  # another process incidentally has the same PID
    raise RuntimeError(f"Another instance of the app is running with PID {prev_process}")


def register_process(name: str):
    own_pid = os.getpid()
    own_process_name = psutil.Process(own_pid).name()
    _check_other_processes(name, own_pid, own_process_name)
    with _pid_file_path(name).open('w') as fd:
        fd.write(f"{own_pid},{own_process_name}")


def unregister_process(name: str):
    pid_file = _pid_file_path(name)
    if not pid_file.exists():
        logger.error(f"PID file missing")
        return
    pid_file.unlink()



import os
import platform
from pathlib import Path


def get_windows_app_dir():
    return Path(os.environ['APPDATA'])


def get_config_folder() -> Path:
    platform_name = platform.system()
    if platform_name == "Darwin":
        folder = Path.home() / "Library" / "Preferences" / "Online Choir"
    elif platform_name == "Windows":
        folder = get_windows_app_dir() / "Online Choir"
    elif platform_name == "Linux":
        folder = Path.home() / ".online-choir"
    else:  # unknown
        return Path(__file__).parent.resolve().parent
    folder.mkdir(exist_ok=True)
    return folder


def get_log_folder() -> Path:
    platform_name = platform.system()
    if platform_name == "Darwin":
        folder = Path.home() / "Library" / "Logs" / "Online Choir"
    elif platform_name == "Windows":
        folder = get_windows_app_dir() / "Online Choir"
    elif platform_name == "Linux":
        folder = Path.home() / ".online-choir"
    else:  # unknown
        return Path(__file__).parent.resolve().parent
    folder.mkdir(exist_ok=True)
    return folder



import audioop
import numpy as np
import struct
from typing import Tuple

from ..common.constants import FRAME_SIZE, BYTE_ENCODING
from ..common.log import logger
from ..common.utils import my_assert

_PRIME_IDX_MULTIPLIER = 199


class AudioEncoder:
    def __init__(self):
        self.next_index = 0

    def __call__(self, sig: np.array) -> bytes:
        return self.create_frame(self._encode_signal(sig))

    def create_frame(self, payload) -> bytes:
        prefix = self._prefix(self.next_index)
        self.next_index += 1
        return prefix + payload

    def _encode_signal(self, sig: np.array) -> bytes:
        my_assert(sig.shape == (FRAME_SIZE,), "wrong signal shape")
        digitalized = sig * 32768
        if digitalized.max() > 32767 or digitalized.min() < -32768:
            logger.warning("Saturation!")
            np.clip(digitalized, -32768, 32767, out=digitalized)
        np.rint(digitalized, out=digitalized)
        return digitalized.astype(BYTE_ENCODING).tobytes()

    @staticmethod
    def _prefix(idx: int):
        return struct.pack("!i", _PRIME_IDX_MULTIPLIER * idx)

    @staticmethod
    def silence_frame(idx: int) -> bytes:
        return AudioEncoder._prefix(idx) + b'\x00' * 1024


class AdpcmEncoder(AudioEncoder):
    def __init__(self):
        super().__init__()
        self._state = None

    def _encode_signal(self, sig: np.array) -> bytes:
        lin = super(AdpcmEncoder, self)._encode_signal(sig)
        comp, self._state = audioop.lin2adpcm(lin, 2, self._state)
        return comp

    @staticmethod
    def silence_frame(idx: int) -> bytes:
        return AudioEncoder._prefix(idx) + b'\x00' * 256


class StereoEncoder(AudioEncoder):
    def __init__(self):
        super().__init__()
        self._state_left = None
        self._state_right = None

    def _encode_signal(self, sig: np.array) -> bytes:
        my_assert(sig.shape == (FRAME_SIZE, 2), "Wrong signal shape")
        left = super(StereoEncoder, self)._encode_signal(sig[:, 0])
        right = super(StereoEncoder, self)._encode_signal(sig[:, 1])
        comp_left, self._state_left = audioop.lin2adpcm(left, 2, self._state_left)
        comp_right, self._state_right = audioop.lin2adpcm(right, 2, self._state_right)
        return comp_left + comp_right

    @staticmethod
    def silence_frame(idx: int) -> bytes:
        return AudioEncoder._prefix(idx) + b'\x00' * 512


class AudioDecoder:
    PACKET_SIZE = 1028

    def __init__(self, next_index: int = 0):
        self.next_index = next_index

    def __call__(self, data: bytes, array: np.ndarray):
        self._check_input(data)
        self._to_array(data, array, 4)

    def _check_input(self, data: bytes):
        my_assert(len(data) == self.PACKET_SIZE, "wrong data length")
        idx = self.peek_frame_index(data)
        my_assert(idx == self.next_index, f"Expected frame {self.next_index} got {idx}")
        self.next_index = idx + 1

    @staticmethod
    def _to_array(data: bytes, array: np.ndarray, offset: int):
        my_assert(array.shape == (FRAME_SIZE,), "wrong signal shape")
        array[:] = np.frombuffer(data, offset=offset, dtype=BYTE_ENCODING) / 32768

    @staticmethod
    def peek_frame_index(data: bytes) -> int:
        n, = struct.unpack("!i", data[:4])
        my_assert(n % _PRIME_IDX_MULTIPLIER == 0, "Invalid data chunk")
        return n // _PRIME_IDX_MULTIPLIER


class AdpcmDecoder(AudioDecoder):
    PACKET_SIZE = 260

    def __init__(self, next_index: int = 0):
        super().__init__(next_index)
        self._state = None

    def __call__(self, data: bytes, array: np.ndarray):
        lin = self.decompress(data, 2)
        self._to_array(lin, array, 0)

    def decompress(self, data: bytes, depth=2) -> bytes:
        self._check_input(data)
        lin, self._state = audioop.adpcm2lin(data[4:], depth, self._state)
        return lin


class StereoDecoder(AudioDecoder):
    PACKET_SIZE = 516

    def __init__(self, next_index: int = 0):
        super().__init__(next_index)
        self._state_left = None
        self._state_right = None

    def __call__(self, data: bytes, array: np.ndarray):
        my_assert(array.shape == (FRAME_SIZE, 2), "wrong signal shape")
        left, right = self.decode_left_right(data)
        array[:, 0] = np.frombuffer(left, dtype=BYTE_ENCODING) / 32768
        array[:, 1] = np.frombuffer(right, dtype=BYTE_ENCODING) / 32768

    def decode_left_right(self, data: bytes, depth=2) -> Tuple[bytes, bytes]:
        self._check_input(data)
        left, self._state_left = audioop.adpcm2lin(data[4:260], depth, self._state_left)
        right, self._state_right = audioop.adpcm2lin(data[260:], depth, self._state_right)
        return left, right



import asyncio
import numpy as np
import queue
import sounddevice as sd
from typing import AsyncIterable, Tuple, Optional

from ..common.audio_devices import get_device_index
from ..common.constants import SAMPLING_RATE, FRAME_SIZE, SLAVE_PLAYBACK_BUFFER_SIZE, FrameType
from ..common.log import logger
from ..common.utils import iterate_queue, flush_queue, my_assert

DEFAULT_ERROR_BUDGET = 5  # number of sounddevice errors to forgive


class AudioIOStream:
    def __init__(self):
        self._input_queue = queue.Queue()
        self._output_queue = asyncio.Queue()
        self._loop = asyncio.get_event_loop()
        self.playing = False
        self._buffer_ready = asyncio.Event()
        self._failed = False
        self._errors_budget = 0

    def _callback(self, indata: np.ndarray, outdata: np.ndarray, frames: int, _, status):
        my_assert(frames == FRAME_SIZE, "wrong frame size")
        if status.output_underflow or status.output_overflow or status.input_overflow:
            logger.error(f"Audio callback error {status} (budget is {self._errors_budget})")
            self._errors_budget -= 1
        try:
            idx, next_chunk = self._input_queue.get_nowait()
        except queue.Empty:
            logger.debug("playback queue empty")
            idx, next_chunk = -2, 0
        outdata[:, :] = next_chunk
        if idx == -1:  # got poison pill
            raise sd.CallbackStop
        if not self._failed:  # record only if not failed
            if self._errors_budget == 0 or idx == -2:
                logger.warning(f"stop recording idx={idx} error budget={self._errors_budget}")
                self._failed = True
                self._put_item_to_output_queue((FrameType.ABORT, None))
                return
            self._put_item_to_output_queue((idx, indata[:, 0].copy()))

    def _finished_callback(self):
        self._put_item_to_output_queue(None)
        logger.debug("Finished callback")

    def _put_item_to_output_queue(self, item):
        self._loop.call_soon_threadsafe(self._output_queue.put_nowait, item)

    async def buffer_ready(self):
        await self._buffer_ready.wait()

    async def play_and_record(self, in_device: str, out_device: str, poll_interval_s: Optional[float] = None) \
            -> AsyncIterable[Tuple[int, np.array]]:
        """
        Plays the content of the input queue (containing data frames as numpy arrays, terminated by a None) as
        returns the corresponding recorded frames. The recorded frames have the same number as the input frames
        (excluding the termination None)
        """
        stream = sd.Stream(device=(get_device_index(in_device, False), get_device_index(out_device, True)),
                           samplerate=SAMPLING_RATE,
                           blocksize=FRAME_SIZE,
                           channels=(1, 2),
                           dtype="float32",
                           callback=self._callback,
                           finished_callback=self._finished_callback,
                           latency='high')
        flush_queue(self._output_queue)
        self._errors_budget = DEFAULT_ERROR_BUDGET
        self._failed = False
        self.playing = True
        with stream:
            async for item in iterate_queue(self._output_queue, poll_interval_s=poll_interval_s):
                if item is None:  # poll interval elapsed, send keep-alive packet
                    yield FrameType.KEEPALIVE, None
                else:
                    yield item
        self.playing = False

    def add_input_chunk(self, idx: int, chunk: np.array):
        assert chunk.shape == (FRAME_SIZE, 2)
        my_assert(idx >= 0, f"Got bad idx={idx}")
        self._input_queue.put_nowait((idx, chunk))
        if not self.playing and self._input_queue.qsize() >= SLAVE_PLAYBACK_BUFFER_SIZE:
            self._buffer_ready.set()

    def add_end_marker(self):
        self._input_queue.put_nowait((-1, None))
        self._buffer_ready.set()  # buffer is ready whenever the audio input is complete

    def flush_input(self):
        my_assert(not self.playing, "flushing while stream is on")
        flush_queue(self._input_queue)
        self._buffer_ready.clear()

    def backlog_size(self) -> int:
        return self._output_queue.qsize()



import asyncio
import platform
import numpy as np
import soundfile as sf
import sounddevice as sd
import tkinter as tk
import threading
import traceback
from functools import partial
from io import StringIO
from queue import Queue
from pathlib import Path
from tkinter.messagebox import showerror
from tkinter.ttk import Separator, Scale, Combobox, Checkbutton, Spinbox, Style
from typing import Optional, Mapping

from .level_meter import BarLevelMeter, DotLevelMeter
from .db_calculator import DbCalculator
from ..common.audio_devices import get_device_list, get_device_index
from ..common.config import default_config, config_path
from ..common.constants import SAMPLING_RATE, FRAME_SIZE, VOCAL_GROUPS
from ..common.log import logger
from ..common.singleton import register_process, unregister_process
from ..common.utils import my_assert, remove_non_ascii
from .calibration_tool import generate_test_audio, MAX_LATENCY, calculate_latency
from .gui_enums import GuiEvent, GuiState
from .signal_smoother import SignalSmoother
from .slave_client import SlaveClient


# TODO:
# * test dead stream
# * handle bad config in sessions


class AudioSettings:
    def __init__(self, parent, destroy, config):
        self.top = tk.Toplevel(parent)
        self.top.title("Settings")
        self.top.protocol("WM_DELETE_WINDOW", destroy)
        self.config = config
        self.level_changes = Queue()
        self.level_updater = threading.Thread(target=self._level_updater, name='level updater')

        row = 0

        # Identity
        tk.Label(self.top, text="Singer Name:").grid(row=row, column=0, sticky='W')
        self.name = tk.Entry(self.top)
        self.name.insert(0, self.config.get('name', ''))
        self.name.grid(row=row, column=1, sticky='EW', padx=10)
        row += 1
        tk.Label(self.top, text="Vocal Group:").grid(row=row, column=0, sticky='W')
        self.vocal_group = Combobox(self.top, values=VOCAL_GROUPS, state='readonly')
        try:
            group_idx = VOCAL_GROUPS.index(self.config.get('group', ''))
        except ValueError:
            group_idx = 0
        self.vocal_group.current(group_idx)
        self.vocal_group.grid(row=row, column=1, sticky='EW', padx=10)
        row += 1

        # Input
        Separator(self.top).grid(row=row, column=0, columnspan=2, sticky='EW', pady=10)
        row += 1
        input_devices, cur_input = get_device_list(self.config, 'audio_input', False)
        tk.Label(self.top, text="Input device:").grid(row=row, column=0, sticky='W')
        self.input_selector = Combobox(self.top, values=input_devices, state='readonly')
        self.input_selector.current(cur_input)
        self.input_selector.bind("<<ComboboxSelected>>", self._switch_input)
        self.input_selector.grid(row=row, column=1, sticky='EW', padx=10)
        row += 1
        tk.Label(self.top, text="Input level:").grid(row=row, column=0, sticky='W')
        self.monitor = BarLevelMeter(self.top, row=row, column=1)
        row += 1
        tk.Label(self.top, text="Microphone Gain:").grid(row=row, column=0, sticky='W')
        gain = config.get('gain', 0)
        self._db_calculator = DbCalculator(gain_db=gain)
        level_scale = Scale(self.top, from_=-20, to=20, command=self._set_gain)
        level_scale.set(gain)
        level_scale.grid(row=row, column=1, sticky='EW', padx=10)
        row += 1

        # Output
        Separator(self.top).grid(row=row, column=0, columnspan=2, sticky='EW', pady=10)
        row += 1
        tk.Label(self.top, text="Output device:").grid(row=row, column=0, sticky='W')
        output_devices, cur_output = get_device_list(self.config, 'audio_output', True)
        self.output_selector = Combobox(self.top, values=output_devices, state='readonly')
        self.output_selector.current(cur_output)
        self.output_selector.bind("<<ComboboxSelected>>", self._switch_output)
        self.output_selector.grid(row=row, column=1, sticky='EW', padx=10)
        row += 1
        self.test = tk.Button(self.top, text="Play Sample Sound", command=self.test_output, state=tk.DISABLED)
        self.test.grid(row=row, column=0, columnspan=2)
        root_folder = Path(__file__).parent.resolve()
        self.sample_sig = np.zeros((345 * FRAME_SIZE, 2))
        try:
            sf.read(root_folder / "data" / "output_test.wav", out=self.sample_sig)
        except RuntimeError:
            msg = "Could not read src/onlinechoir/slave/data/output_test.wav. " \
                  "Did you set up git-lfs and checked it out correctly?"
            logger.error(msg)
            raise RuntimeError(msg)
        row += 1

        # Calibration
        Separator(self.top).grid(row=row, column=0, columnspan=2, sticky='EW', pady=10)
        row += 1
        Style().configure("Red.TSpinbox", background="red")
        tk.Label(self.top, text="Current latency:").grid(row=row, column=0, sticky='W')
        self.latency_box = Spinbox(self.top, increment=1, from_=0, to=200, command=self._latency_box_callback, width=4)
        self._set_latency(int(config['latency'] or 0))
        self.latency_box.grid(row=row, column=1, sticky='W', padx=10)
        row += 1
        tk.Label(self.top, text="A latency value of 0 means that the latency is undefined yet.\n"
                                "To measure the latency, place your headphones close\n"
                                "to the microphone, raise the volume and click below.\n"
                                "Do not change the latency value manually unless you know what you are doing!") \
            .grid(row=row, column=0, columnspan=2)
        row += 1
        self.calib = tk.Button(self.top, text="Measure Latency", command=self.calibrate, fg="red", state=tk.DISABLED)
        self.calib.grid(row=row, column=0, columnspan=2)
        row += 1
        tk.Label(self.top, text="Do not press the above button with the headphones on your ears!", fg="red") \
            .grid(row=row, column=0, columnspan=2)
        weep = generate_test_audio()
        self.weep_len, n_chan = weep.shape
        my_assert(n_chan == 2, "Wrong weep shape")
        self.out_signal = np.zeros((self.weep_len + FRAME_SIZE * MAX_LATENCY, 2), dtype=np.float32)
        self.out_signal[:self.weep_len, :] = weep
        self.in_signal = np.empty((self.weep_len + FRAME_SIZE * MAX_LATENCY,), dtype=np.float32)
        row += 1

        # Server selection
        Separator(self.top).grid(row=row, column=0, columnspan=2, sticky='EW', pady=10)
        row += 1
        tk.Label(self.top, text="Server address:").grid(row=row, column=0, sticky='W')
        self.server_address = tk.Entry(self.top, width=20)
        self.server_address.insert(0, self.config['server'])
        self.server_address.grid(row=row, column=1, sticky='W', padx=10)
        row += 1
        tk.Label(self.top, text="Server port:").grid(row=row, column=0, sticky='W')
        self.server_port = tk.Entry(self.top, width=5)
        self.server_port.insert(0, self.config['port'])
        self.server_port.grid(row=row, column=1, sticky='W', padx=10)
        row += 1

        # Close button
        Separator(self.top).grid(row=row, column=0, columnspan=2, sticky='EW', pady=10)
        row += 1
        self.ok = tk.Button(self.top, text="OK", command=destroy)
        self.ok.grid(row=row, column=0, columnspan=2)
        row += 1

        # Start audio stream
        self.mutex = threading.Lock()
        self.cursor = None
        self.rec = None
        self.stream = None
        self._start_stream(self.config.get('audio_input'), self.config.get('audio_output'))
        self.level_updater.start()

    def _start_stream(self, in_device: str, out_device: str):
        my_assert(self.stream is None, "Stream should be None")
        devices = (get_device_index(in_device, False), get_device_index(out_device, True))
        logger.debug(f"devices are {devices}")
        try:
            self.stream = sd.Stream(device=devices,
                                    samplerate=SAMPLING_RATE,
                                    blocksize=FRAME_SIZE,
                                    channels=(1, 2),
                                    dtype="float32",
                                    callback=self._player_callback,
                                    latency='high')
            self.stream.start()
        except (sd.PortAudioError, ValueError) as e:
            logger.error(f"Failed to start stream: {e}")
            return

    def _stop_stream(self):
        if self.stream is not None:
            self.test.config(state=tk.DISABLED)
            self.calib.config(state=tk.DISABLED)
            self.stream.close()
            self.level_changes.put_nowait(0)
            self.stream = None

    def _player_callback(self, indata: np.ndarray, outdata: np.ndarray, frames: int, _, status):
        # General switch: self.rec, self.cursor
        # None, None: idle
        # False, 0: request to play sample
        # True, 0: request to calibrate
        # False/True, >0: play
        # False/True, -1: process
        my_assert(frames == FRAME_SIZE, "wrong frame size")
        with self.mutex:
            rec = self.rec
            cursor = self.cursor
        if status:  # error case
            if rec and cursor >= 0:
                logger.error(f"callback error {status}, calibration failed")
                threading.Thread(target=self._processing, name='failed processing').start()  # cancel
            else:
                logger.debug(f"Got status {status}")
            return  # do nothing else

        if self.level_changes.empty():  # skip if updater is behind
            self.level_changes.put_nowait(indata.copy())

        if rec is None or cursor < 0:  # play silence, ignore input
            outdata[:] = 0.
            return

        my_assert(cursor is not None, "cursor not be None when rec is not None")
        sig = self.out_signal if self.rec else self.sample_sig
        outdata[:, :] = sig[self.cursor: self.cursor + FRAME_SIZE, :]
        if self.rec:  # is True
            self.in_signal[self.cursor: self.cursor + FRAME_SIZE] = indata[:, 0]
        self.cursor += FRAME_SIZE
        if self.cursor >= len(sig):
            logger.debug("Finished playing, start processor")
            self.cursor = -1
            threading.Thread(target=self._processing, name='processing').start()  # process the result

    def _set_latency(self, latency: int):
        self.latency_box.set(latency)
        self._update_latency_box_color(latency)

    def _latency_box_callback(self):
        self._update_latency_box_color(int(self.latency_box.get()))

    def _update_latency_box_color(self, value: int):
        if value > 0:
            self.latency_box.configure(style="TSpinbox")
        else:
            self.latency_box.configure(style="Red.TSpinbox")

    def _processing(self):
        my_assert(self.cursor is not None, "This thread should not start if cursor is None")
        my_assert(self.rec is not None, "This thread should not start if rec is None")
        if self.rec:
            if self.cursor == -1:  # recording complete
                latency = calculate_latency(self.out_signal[:self.weep_len].mean(axis=1),
                                            self.in_signal * self._db_calculator.gain)
                logger.info(f"Processing done. Latency is {latency or 'not measured successfully'}")
                self._set_latency(int(round(latency or 0)))
            else:  # recording incomplete
                logger.warning("Recording interrupted.")
        with self.mutex:
            self.rec = None
            self.cursor = None
        self.calib.config(state=tk.NORMAL)
        self.test.config(state=tk.NORMAL)
        self.ok.config(state=tk.NORMAL)
        self.latency_box.configure(state='normal')
        self.input_selector.configure(state='readonly')
        self.output_selector.configure(state='readonly')
        logger.debug("Processing completed")

    def _level_updater(self):
        logger.debug("Start level updater")
        stream_active = False
        while True:
            sample = self.level_changes.get()
            if sample is None:  # poison pill
                break
            if isinstance(sample, int):  # stream ended
                self.monitor.reset()
                stream_active = False
            else:  # actual numpy array
                if not stream_active:
                    self._stream_active()
                    stream_active = True
                db = self._db_calculator.from_sample(sample)
                self.monitor.update_level(db)
        logger.debug("End level updater")

    def _stream_active(self):
        self.test.config(state=tk.NORMAL)
        self.calib.config(state=tk.NORMAL)
        self.config['audio_input'] = remove_non_ascii(self.input_selector.get())
        self.config['audio_output'] = remove_non_ascii(self.output_selector.get())

    def _set_gain(self, gain: str):  # gain is in dB and is a string because it is a tkinter widget
        self.config['gain'] = gain
        self._db_calculator.set_gain(gain_db=gain)
        logger.debug(f"Gain set to {self._db_calculator.gain} ({self._db_calculator.gain_db} dB)")

    def test_output(self):
        logger.info("Testing output")
        self._busy()
        with self.mutex:
            self.rec = False
            self.cursor = 0

    def calibrate(self):
        logger.info("Calibrate")
        self._busy()
        with self.mutex:
            self.rec = True
            self.cursor = 0

    def _busy(self):
        self.calib.config(state=tk.DISABLED)
        self.test.config(state=tk.DISABLED)
        self.ok.config(state=tk.DISABLED)
        self.latency_box.configure(state='disabled')
        self.input_selector.configure(state='disabled')
        self.output_selector.configure(state='disabled')

    def _switch_input(self, _):
        new_input = self.input_selector.get()
        logger.info(f"Changing audio input to {remove_non_ascii(new_input)}")
        self._restart_audio(new_input, self.config.get('audio_output', ''))

    def _switch_output(self, _):
        new_output = self.output_selector.get()
        logger.info(f"Changing audio output to {remove_non_ascii(new_output)}")
        self._restart_audio(self.config.get('audio_input', ''), new_output)

    def _restart_audio(self, in_device: str, out_device: str):
        self._set_latency(0)
        logger.debug("Restarting audio stream")
        if self.rec is not None:
            logger.error("Cannot restart audio stream, as it is playing")
            return
        self._stop_stream()
        self._start_stream(in_device, out_device)

    def _get_name(self) -> str:
        lines = self.name.get().splitlines()
        if lines:
            return lines[0].strip()
        return ''

    def destroy(self):
        with self.mutex:
            playing = self.rec is not None
        if playing:
            return False
        self.level_changes.put_nowait(None)
        self.level_updater.join()
        self._stop_stream()
        self.config['server'] = self.server_address.get()
        self.config['port'] = self.server_port.get()
        self.config['name'] = self._get_name()
        self.config['group'] = self.vocal_group.get()
        self.config['latency'] = self.latency_box.get()
        self.top.destroy()
        return True


class SlaveClientGui:
    def __init__(self, save_config=None):
        self.events_out: Optional[asyncio.Queue] = None
        self.loop: Optional[asyncio.AbstractEventLoop] = None
        self.save_config = save_config

        # create window
        self.root = tk.Tk()
        self.root.title(f"Online Choir")
        self.label = tk.Label(master=self.root, text="CONNECTING...")
        self.label.grid(row=1, column=1, padx=10, pady=10)

        self.mute = tk.IntVar(self.root, 1, 'mute')
        Checkbutton(self.root, text="Mute", variable=self.mute,
                    offvalue=1, onvalue=0).grid(row=2, column=1, padx=10, pady=10)
        self.button = tk.Button(master=self.root, text="Settings", width=20, state=tk.DISABLED,
                                command=partial(self.send_event, GuiEvent.SETTINGS_BUTTON.value))
        self.button.grid(row=3, column=1, columnspan=2, padx=10, pady=10)

        self.monitor = DotLevelMeter(self.root, row=1, column=2, padx=10, pady=10)

        # State variables
        self.state = None
        self.signal_smoother = SignalSmoother(-1)

        # Settings
        self.settings = None

    def spin(self):
        try:
            self.root.mainloop()
        finally:
            self.state = GuiState.DEAD  # nullify state after the GUI has stopped
            if self.loop.is_running():
                logger.debug("Sending null event to business logic")
                self.send_event(None)

    # this is to be called from another thread
    def change_state(self, state: int, message: Optional[str] = None):
        if self.state == GuiState.DEAD:
            return  # ignore post-mortem commands
        state = GuiState(state)
        my_assert(state != self.state, "repeated state")
        logger.info(f"Changing state to {state.name}")
        self.state = state
        text = state.name
        if message:
            text += f"({message})"
        self.label["text"] = text
        if state in (GuiState.READY, GuiState.DISCONNECTED):
            self.button.config(state=tk.NORMAL)
        else:
            self.button.config(state=tk.DISABLED)

    def send_event(self, event: Optional[int]):
        self.loop.call_soon_threadsafe(self.events_out.put_nowait, event)

    # this is to be called from another thread
    def set_signal_level(self, db: Optional[float]):
        if self.state == GuiState.DEAD:
            return  # ignore post-mortem commands
        self.monitor.update_level(db)

    def open_settings(self, config: Mapping[str, str]):
        self.settings = AudioSettings(self.root, self.close_settings, config)

    def close_settings(self):
        logger.debug("close settings")
        if self.settings.destroy():
            if self.save_config:
                self.save_config()
            self.send_event(GuiEvent.GUI_READY.value)
            self.settings = None
            return True
        return False


def main():
    logger.info(f"Started on {platform.system()}")
    try:
        register_process('singer_gui')
    except RuntimeError:
        showerror(message='Another instance of Online Choir is running')
        return
    config = default_config()
    if 'slave' not in config.sections():
        config['slave'] = {'server': 'online-choir.modetexte.ch',
                           'port': '8868',
                           'latency': '0'}
    logger.debug(f"Loaded config. Latency is {config['slave']['latency']}.")

    def save_config():
        logger.debug("Saving config")
        with config_path().open('w') as fd:
            config.write(fd)

    gui = SlaveClientGui(save_config)
    client = SlaveClient(config['slave'], gui)

    def run_business_logic():
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(client.main())
        except Exception as e:
            logger.error(f"Business logic crashed with error {e}")
            sf = StringIO()
            traceback.print_tb(e.__traceback__, file=sf)
            logger.debug(sf.getvalue())
            gui.change_state(GuiState.DISCONNECTED.value)
        finally:
            loop.close()
        return

    business_logic = threading.Thread(target=run_business_logic, name='business logic')
    logger.debug("Start business logic thread")
    business_logic.start()
    logger.debug("Start GUI")
    gui.spin()
    logger.debug("Wait for business logic to exit")
    business_logic.join()
    unregister_process('singer_gui')
    logger.debug("Done")



def app():
    """Entry point for the packaged (pyinstaller) app"""
    from onlinechoir.common.log import setup_logging
    from onlinechoir.common.platform import get_log_folder
    from .gui import main
    setup_logging('ERROR', get_log_folder() / "slave_gui.log")
    main()


def cli():
    """Entry point for the pure python slave client"""
    import argparse

    from onlinechoir.common.log import parse_args_and_set_logger
    from .slave_client import main as main_cli
    from .gui import main as main_gui

    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--no-gui', action="store_true",
                        help="Run the command-line client")
    parser.add_argument('-s', '--server', default='localhost',
                        help="(CLI only) The address of the server")
    parser.add_argument('-p', '--port', type=int, default=8868,
                        help="(CLI only) The port number to be used on the server (slave port)")
    parser.add_argument('-l', '--latency', type=int, default=5,
                        help="(CLI only) The latency of the audio hardware (to be measured separately)")

    parsed_args = parse_args_and_set_logger(parser, 'slave')

    if parsed_args.no_gui:
        main_cli(server=parsed_args.server, port=parsed_args.port, latency=parsed_args.latency)
    else:
        main_gui()



import sounddevice as sd
from typing import Tuple, List, MutableMapping, Optional

from .constants import SAMPLING_RATE
from .log import logger
from .utils import my_assert, remove_non_ascii

try:
    HOST_API = sd.query_devices(kind='input')['hostapi']  # freeze host API at start
except sd.PortAudioError:
    HOST_API = -1  # only for unit tests


def get_device_list(config: MutableMapping[str, str], entry: str, out: bool) -> Tuple[List[str], int]:
    """
    Returns the list of available devices and the index of the currently selected device (from provided config)
    :param config: current config
    :param entry: relevant entry of the config
    :param out: whether we want out-devices or in-devices
    """
    config_device = config.get(entry, '')
    field = 'max_output_channels' if out else 'max_input_channels'
    chans = 2 if out else 1
    default_idx = sd.query_hostapis(HOST_API)['default_output_device' if out else 'default_input_device']
    devices = []
    selected = None
    default = None
    default_name = None
    for idx, device in enumerate(sd.query_devices()):
        if device['hostapi'] != HOST_API:
            continue
        if device[field] < chans:
            continue
        name = device['name'].strip()
        devices.append(name)
        sanitized_name = remove_non_ascii(name)
        if sanitized_name == config_device:
            selected = len(devices) - 1
        if idx == default_idx:
            default = len(devices) - 1
            default_name = name
    my_assert(len(devices) > 0, f"No suitable device found for {'output' if out else 'input'}.")  # TODO: improve
    if default is None:
        logger.warning("Default device not suitable; using first of the list")
        default = 0
        default_name = devices[0]
    if selected is None:
        logger.warning(f"Device from config ({remove_non_ascii(config_device)}) not in device list; switch to default")
        selected = default
        config[entry] = remove_non_ascii(default_name)
        if 'latency' in config:  # reset latency
            config['latency'] = '0'
    return devices, selected


def check_audio_config(config: MutableMapping[str, str], in_entry: str, out_entry: str):
    if HOST_API < 0:  # skip this when testing
        return
    logger.debug(f"Checking audio config. Host API is {HOST_API}")
    # fix config if needed
    get_device_list(config, in_entry, False)
    get_device_list(config, out_entry, True)
    try:
        sd.check_input_settings(device=get_device_index(config.get(in_entry, ''), False),
                                channels=1,
                                dtype='float32',
                                samplerate=SAMPLING_RATE)
        sd.check_output_settings(device=get_device_index(config.get(out_entry, ''), True),
                                 channels=2,
                                 dtype='float32',
                                 samplerate=SAMPLING_RATE)
    except sd.PortAudioError:
        logger.warning(f"Audio config invalid; resetting latency")
        if 'latency' in config:  # reset latency, preventing client to join
            config['latency'] = '0'


def get_device_index(selected_name: str, out: bool) -> Optional[int]:
    selected_name = remove_non_ascii(selected_name.strip())
    field = 'max_output_channels' if out else 'max_input_channels'
    chans = 2 if out else 1
    for i, dev in enumerate(sd.query_devices()):
        sanizized_name = remove_non_ascii(dev['name'].strip())
        if sanizized_name == selected_name and dev['hostapi'] == HOST_API and dev[field] >= chans:
            return i
    logger.error(f"Bad device name {remove_non_ascii(selected_name)}, using default")
    return None



import asyncio
import queue
from math import log10
from typing import Union, AsyncIterable, Optional, TypeVar, Iterable, Callable

from .log import logger


async def iterate_queue(q: asyncio.Queue, poll_interval_s: Optional[float] = None, sentry=None) -> AsyncIterable:
    """Iterates over the elements of the queue, waiting for next elements if necessary, until 'sentry' is
       retrieved. If poll_interval_s is not None, also yields a None periodically."""
    while True:
        try:
            item = await asyncio.wait_for(q.get(), poll_interval_s)
        except asyncio.TimeoutError:
            yield None
            continue
        if item is sentry:
            q.task_done()
            return
        yield item
        q.task_done()


def flush_queue(q: Union[asyncio.Queue, queue.Queue]) -> int:
    n = 0
    while not q.empty():
        q.get_nowait()
        q.task_done()
        n += 1
    return n


async def iterate_stream(stream: asyncio.StreamReader, chunk_size: int, timeout: Optional[float] = None) \
        -> AsyncIterable:
    while True:
        try:
            data = await asyncio.wait_for(stream.readexactly(chunk_size), timeout)
        except asyncio.IncompleteReadError:
            data = None
        except ConnectionResetError:
            logger.warning("Connection reset")
            data = None
        if not data:
            return
        yield data


def my_assert(condition: bool, message: str):
    if not condition:
        logger.error(message)
        raise RuntimeError(message)


def to_db(gain) -> float:
    return 20 * log10(float(gain))


def to_gain(db) -> float:
    return 10 ** (float(db) / 20)


def remove_non_ascii(s) -> str:
    return str(s).encode('ascii', errors='replace').decode()


Item = TypeVar("Item")


def find_first(iterable: Iterable[Item], condition: Callable[[Item], bool], default: Optional[Item] = None) -> Item:
    for item in iterable:
        if condition(item):
            return item
    if default is None:
        raise IndexError("no item found")
    return default



from enum import Enum


class GuiEvent(Enum):
    SESSION_STARTS = 1
    SETTINGS_BUTTON = 2
    GUI_READY = 3


class GuiState(Enum):
    CONNECTING = 0
    READY = 1
    PLAYING = 2
    SETTINGS = 3
    DISCONNECTED = 4
    DEAD = 5
    ERROR = 6



from typing import Optional, Union

import numpy as np

from ..common.constants import SATURATION_LEVEL
from ..common.utils import to_db, to_gain


class DbCalculator:
    """
    Calculates the signal level in dBs.
    By convention, the following levels are defined:
    * 0 dB: the maximum before saturation
    * 10 dB: saturation
    * -100 dB: perfectly silent (but signal present)
    """
    def __init__(self, *, gain: Optional[Union[float, str]] = None, gain_db: Optional[Union[float, str]] = None):
        self.set_gain(gain=gain, gain_db=gain_db)

    def set_gain(self, *, gain: Optional[Union[float, str]] = None, gain_db: Optional[Union[float, str]] = None):
        if gain is not None:
            if gain_db is not None:
                raise ValueError("Cannot specify both gain and gain_db")
            self.gain, self.gain_db = float(gain), to_db(gain)
        elif gain_db is not None:
            self.gain, self.gain_db = to_gain(gain_db), float(gain_db)
        else:
            raise ValueError("Must specify either gain or gain_db")

    def from_sample(self, sample: np.ndarray) -> float:
        peak = np.abs(sample).max()
        if peak * self.gain >= SATURATION_LEVEL:
            return 10.
        var = sample.var()  # signal can clip if var > 0.1111, therefore we set the 0dB to 0.1111.
        if var == 0:
            return -100.
        return 10 * np.log10(9 * var) + self.gain_db

    def from_chunk(self, chunk: bytes) -> float:
        pass



import asyncio
import sys


class Prompt:
    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.q = asyncio.Queue(loop=self.loop)
        self.loop.add_reader(sys.stdin, self.got_input)

    def got_input(self):
        asyncio.ensure_future(self.q.put(sys.stdin.readline()), loop=self.loop)

    async def __call__(self, msg, end='\n', flush=False):
        print(msg, end=end, flush=flush)
        return (await self.q.get()).rstrip('\n')



"""
The client program to be used by signers.
"""

import asyncio
import numpy as np
from io import StringIO
from typing import Optional, MutableMapping, Any

from .db_calculator import DbCalculator
from ..common.audio_codecs import StereoDecoder, AdpcmEncoder
from ..common.audio_devices import check_audio_config
from ..common.constants import FRAME_SIZE, FrameType, KEEPALIVE_INTERVAL, CONNECTION_TIMEOUT
from ..common.version import VERSION
from ..common.auth import AuthenticationHeader
from ..common.log import logger
from ..common.network import set_socket_settings
from ..common.utils import iterate_queue, iterate_stream, my_assert, to_gain
from .gui_enums import GuiEvent, GuiState
from .sound_card import AudioIOStream


# Pipe structure:
# * TCP in: frames 0:n, ended by empty frame -1
# * TCP out: frames 0:n, started with -1, possibly ended by -2 (if ended early)


class SlaveClient:
    def __init__(self, in_config: MutableMapping[str, Any], gui, skip_check=False):
        self._config = in_config
        self._gui = gui
        self._state = None
        self._state_changed = None
        self._audio = None
        self._events = None
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._receive_task = None

        if not skip_check:
            check_audio_config(self._config, 'audio_input', 'audio_output')

        self._latency = int(in_config['latency'])

    def _change_state(self, state: GuiState, message: Optional[str] = None):
        self._state = state
        if self._gui:
            self._gui.change_state(state.value, message)
        self._state_changed.set()

    async def _receive_audio(self):
        # Keep decoding input TCP, queue it only if in the right state.
        # Keeps account of sections.
        # Adds the silence frames needed to compensate for latency.
        decoder: Optional[StereoDecoder] = None

        def start_session():
            nonlocal decoder
            logger.debug("Init decoder, flush audio queue and join session")
            self._events.put_nowait(GuiEvent.SESSION_STARTS.value)
            decoder = StereoDecoder()
            self._audio.flush_input()

        try:
            async for data in iterate_stream(self._reader, StereoDecoder.PACKET_SIZE, timeout=CONNECTION_TIMEOUT):
                idx = StereoDecoder.peek_frame_index(data)
                if idx == FrameType.KEEPALIVE:
                    logger.debug("got keep-alive message form server")
                    continue
                if idx == FrameType.START:
                    logger.debug("Session starting")
                    my_assert(decoder is None, "idx = 0 while decoder exists")
                    if self._latency == 0:
                        logger.warning("Session started, but client is not calibrated. Skipping.")
                        continue
                    if self._state == GuiState.READY:
                        start_session()
                    elif self._state == GuiState.SETTINGS and self._gui is not None:
                        logger.debug("settings still open, trying to close them")
                        self._state_changed.clear()
                        if self._gui.close_settings():
                            await self._state_changed.wait()
                            my_assert(self._state == GuiState.READY, "State should be READY here")
                            start_session()
                        else:
                            logger.warning("Session starting but settings can not be closed")
                    else:
                        logger.error(f"Session starting but state is {self._state}")
                if decoder is not None:
                    if self._state not in (GuiState.READY, GuiState.PLAYING):
                        logger.warning("State changed during session. Interrupting")
                        decoder = None
                        continue
                    if idx >= 0:
                        sig = np.empty((FRAME_SIZE, 2), dtype=np.float32)
                        decoder(data, sig)
                        self._audio.add_input_chunk(idx, sig)
                    else:
                        my_assert(idx == FrameType.EOF, f"Expected audio frame or EOF, but got {idx}")
                        logger.debug("Session ending")
                        # add silence to compensate for latency
                        silence = np.zeros((FRAME_SIZE, 2), dtype=np.float32)
                        next_idx = decoder.next_index
                        for i in range(next_idx, next_idx + self._latency):
                            self._audio.add_input_chunk(i, silence)
                        self._audio.add_end_marker()
                        decoder = None
        except asyncio.TimeoutError:
            logger.warning("Server not responding")
        except ConnectionResetError:
            logger.warning("Connection got reset")
        except asyncio.CancelledError:
            logger.info("Receiver cancelled")
        else:
            logger.info("Server closed connection")
        if decoder:  # stop playing if connection is lost
            logger.warning("Add end marker since server closed during session")
            self._audio.add_end_marker()

    async def _record_session(self):
        my_assert(self._state == GuiState.READY, "start playing while state is not ready")
        self._change_state(GuiState.PLAYING)
        self._writer.write(AdpcmEncoder.silence_frame(FrameType.CHECK_IN))
        await self._writer.drain()
        await self._audio.buffer_ready()
        logger.info("Sing!")
        latency_drop_count = self._latency
        db_calculator = DbCalculator(gain_db=self._config.get('gain', 0))
        encoder = AdpcmEncoder()
        failed = False
        async for idx, chunk in self._audio.play_and_record(self._config.get('audio_input', ''),
                                                            self._config.get('audio_output', ''),
                                                            KEEPALIVE_INTERVAL):
            if idx == FrameType.ABORT:
                logger.warning("Session failed. Dropping")
                data = AdpcmEncoder.silence_frame(FrameType.ABORT)
                failed = True
                if self._gui is not None:
                    self._gui.set_signal_level(None)
            elif idx == FrameType.KEEPALIVE:
                data = AdpcmEncoder.silence_frame(FrameType.KEEPALIVE)
            else:
                my_assert(not failed, "Audio came after the drop frame")
                # report on signal level if the system is not lagging
                if self._gui is not None and self._audio.backlog_size() == 0:
                    db = db_calculator.from_sample(chunk)
                    self._gui.set_signal_level(db)
                if latency_drop_count:
                    latency_drop_count -= 1
                    continue
                else:
                    my_assert(idx == encoder.next_index + self._latency, "inconsistent chunks")
                    data = encoder((db_calculator.gain * (1 if self._gui is None else self._gui.mute.get())) * chunk)
            self._writer.write(data)
            if not self._writer.is_closing():
                await self._writer.drain()
        if failed:
            logger.info("Thanks! It did not go well...")
        else:
            logger.info("Beautiful, thanks!")
            if self._gui is not None:
                self._gui.set_signal_level(None)
        if self._receive_task.done():  # lost connection to server
            self._change_state(GuiState.DISCONNECTED)
        else:
            self._change_state(GuiState.READY)

    async def _open_settings(self):
        my_assert(self._gui is not None, "settings button pressed without a GUI...")
        self._change_state(GuiState.SETTINGS)  # after this, the receiver is paused.
        server_settings = self._config['server'], str(self._config['port']), \
            self._config.get('name'), self._config.get('group')
        self._gui.open_settings(self._config)
        if self._writer is not None and not self._writer.is_closing():
            logger.debug("send initial keep-alive to server from open settings")
            self._writer.write(AdpcmEncoder.silence_frame(FrameType.KEEPALIVE))
        event = None
        while event is None:
            try:
                event = await asyncio.wait_for(self._events.get(), timeout=KEEPALIVE_INTERVAL)
            except asyncio.TimeoutError:
                if self._writer is not None and not self._writer.is_closing():
                    logger.debug("send keep-alive to server from open settings")
                    self._writer.write(AdpcmEncoder.silence_frame(FrameType.KEEPALIVE))
        my_assert(event == GuiEvent.GUI_READY.value, f"Unexpected event {event}")
        self._latency = int(self._config['latency'])
        logger.debug(f"got GUI ready event; settings closed, latency is {self._latency}, "
                     f"gain is {to_gain(self._config.get('gain', 0))}")
        if server_settings != (self._config['server'], str(self._config['port']),
                               self._config.get('name'), self._config.get('group')):
            logger.info(f"New server settings: {self._config['server']}:{self._config['port']}")
            if self._receive_task is not None and not self._receive_task.done():
                self._receive_task.cancel()
                self._writer.close()
            self._change_state(GuiState.CONNECTING)
            await self._connect_to_server()
        elif self._receive_task is None or self._receive_task.done():  # attempt a reconnection
            self._change_state(GuiState.CONNECTING)
            await self._connect_to_server()
        else:
            self._change_state(GuiState.READY, "Uncalibrated!" if self._latency == 0 else None)

    async def _connect_to_server(self) -> bool:
        # Connect to server
        try:
            self._reader, self._writer = await asyncio.open_connection(self._config['server'],
                                                                       int(self._config['port']))
        except ValueError as e:
            logger.error(f"Bad server settings {e}")
            self._change_state(GuiState.DISCONNECTED)
            return False
        except OSError as e:
            logger.warning(f"Connection failed with error {e}")
            self._change_state(GuiState.DISCONNECTED)
            return False
        logger.info(f"Connected to server {self._config['server']}:{self._config['port']} "
                    f"from {self._writer.get_extra_info('sockname')}")
        set_socket_settings(self._writer)

        # Authenticate
        self._writer.write(AuthenticationHeader(self._config.get('name', ''), self._config.get('group', '')).to_bytes())
        await self._writer.drain()

        # Start receiver
        self._change_state(GuiState.READY, "Uncalibrated!" if self._latency == 0 else None)
        self._receive_task = asyncio.create_task(self._receive_audio())
        return True

    async def main(self):
        self._audio = AudioIOStream()
        self._events = asyncio.Queue()
        self._state_changed = asyncio.Event()
        if self._gui:
            self._gui.events_out = self._events
            self._gui.loop = asyncio.get_event_loop()
        logger.info(f"Starting slave client v{VERSION}")
        await self._connect_to_server()

        # Event loop
        logger.debug("Start waiting for events")
        async for event in iterate_queue(self._events, poll_interval_s=KEEPALIVE_INTERVAL):
            if event is not None:
                assert isinstance(event, int)
                event = GuiEvent(event)

                if event == GuiEvent.SESSION_STARTS:
                    logger.debug("Got session start event")
                    await self._record_session()

                elif event == GuiEvent.SETTINGS_BUTTON:
                    if self._state not in (GuiState.READY, GuiState.DISCONNECTED):
                        continue  # skip button presses that are not in the right state
                    logger.debug("Got settings button event")
                    await self._open_settings()

                else:
                    logger.error(f"Not implemented: {event}")

            # check if receive_task is still alive
            if self._state != GuiState.READY:
                continue  # don't check if we are disconnected
            try:
                receive_exception = self._receive_task.exception()
            except asyncio.InvalidStateError:  # up and running, yeay!
                logger.debug("send keep-alive to server")
                self._writer.write(AdpcmEncoder.silence_frame(FrameType.KEEPALIVE))
                continue
            if receive_exception is not None:
                f = StringIO()
                self._receive_task.print_stack(file=f)
                logger.error(f"Receive task died with exception {receive_exception}, traceback {f.getvalue()}")
                self._change_state(GuiState.ERROR)
                break
            else:
                logger.warning("Got disconnected")
                self._change_state(GuiState.DISCONNECTED)

        # clean up before exit
        if self._writer is not None:
            self._writer.close()
        if self._receive_task is not None:
            logger.debug("waiting for receiver to complete")
            await self._receive_task
        return


def main(*, server: str, port: int, latency: int):
    config = {'server': server, 'port': port, 'latency': latency}
    client = SlaveClient(config, None)
    asyncio.run(client.main())
    exit(0)



import logging
import sys
from logging.handlers import RotatingFileHandler
from typing import Optional

from .platform import get_log_folder

logger = logging.getLogger('online-choir')


class StreamToLogger:
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, output_logger, level=logging.DEBUG):
        self.logger = output_logger
        self.log_level = level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):
        pass


def setup_logging(log_level, log_file=None, capture_stderr=True):
    formatter = logging.Formatter('%(asctime)s:%(levelname)s:%(threadName)s:%(message)s')
    logger.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    if log_file:
        logger.info(f"Logging output to {log_file}")
        file_handler = RotatingFileHandler(log_file, maxBytes=10 * 2**20, backupCount=5)
        file_handler.setFormatter(formatter)
        logging.root.addHandler(file_handler)

        if capture_stderr:
            # send stderr to logger (only if writing to a file)
            stderr_logger = logging.getLogger('STDERR')
            sl = StreamToLogger(stderr_logger, logging.ERROR)
            sys.stderr = sl


def parse_args_and_set_logger(parser, program: Optional[str] = None):
    parser.add_argument('--log-level', choices=('DEBUG', 'INFO', 'WARNING', 'ERROR'), default='INFO')
    parser.add_argument('--log-file')
    parser.add_argument('--console', action="store_true", help="Do not redirect stderr to log file")
    parsed_args = parser.parse_args()
    log_file = parsed_args.log_file
    if log_file is None and program:
        # enable logging in default location
        if parsed_args.console:
            log_file = f"{program}.log"
        else:
            log_file = get_log_folder() / f"{program}.log"
    setup_logging(parsed_args.log_level, log_file, not parsed_args.console)
    return parsed_args



import audioop
import numpy as np
import soundfile
import subprocess
from pathlib import Path

from .constants import SAMPLING_RATE, FRAME_SIZE, BYTE_ENCODING
from .log import logger
from .utils import my_assert

# Initialize path to faad
FAAD_PATH1 = Path('/usr/local/bin/faad')
FAAD_PATH2 = Path('/usr/bin/faad')
if FAAD_PATH1.exists():
    FAAD_PATH = FAAD_PATH1
elif FAAD_PATH2.exists():
    FAAD_PATH = FAAD_PATH2
else:
    FAAD_PATH = None


def load_audio_file(file_name) -> np.ndarray:
    with soundfile.SoundFile(file_name) as f:
        if f.samplerate != SAMPLING_RATE:
            raise ValueError("Wrong sampling rate")
        logger.debug(f"File has {f.frames} samples and {f.channels} channels")
        my_assert(f.channels <= 2, "Too many channels")
        n_chunks = ((f.frames - 1) // FRAME_SIZE) + 1
        res = np.zeros((n_chunks * FRAME_SIZE, f.channels), dtype='float64', order='C')
        assert res.shape[0] >= f.frames
        to_read = f.frames
        read = f.read(to_read, out=res)
        # Workaround
        while len(read) == 0:  # failed to read
            logger.debug("failed read, chopping a bit of the file")
            to_read -= 4096
            if to_read <= 0:
                raise RuntimeError("Failed to read file with soundfile")
            read = f.read(to_read, out=res)
    if res.shape[1] == 1:
        res = np.outer(res, np.ones((2,)))
    assert res.shape == (n_chunks * FRAME_SIZE, 2)
    logger.info(f"Loaded {n_chunks} chunks")
    return res


def is_m4a_supported() -> bool:
    return FAAD_PATH is not None


def load_m4a_file(file_name: str) -> np.ndarray:
    if FAAD_PATH is None:
        raise RuntimeError("FAAD missing")
    res = subprocess.run([FAAD_PATH, "-b", "1", "-f", "2", "-w", file_name], capture_output=True)
    if res.returncode != 0:
        raise ValueError("Failed to decode file")

    # parse file info
    rate = None
    channels = 0
    for line in res.stderr.decode('UTF8').splitlines():
        if line.startswith('Samplerate:'):
            rate = int(line.split('\t')[-1])
        elif line.startswith('Total channels:'):
            channels = int(line.split('\t')[-1])
    if rate != SAMPLING_RATE:
        raise ValueError("Wrong sampling rate")
    if not 1 <= channels <= 2:
        raise ValueError("Wrong number of channels")

    # reformat audio output
    data = res.stdout
    if channels == 2:
        left = np.frombuffer(audioop.tomono(data, 2, 1, 0), dtype='i2')
        right = np.frombuffer(audioop.tomono(data, 2, 0, 1), dtype='i2')
    else:
        left = np.frombuffer(data, dtype='i2')
        right = left
    assert len(left) == len(right)
    n_frames = len(left)
    n_chunks = ((n_frames - 1) // FRAME_SIZE) + 1
    sig = np.zeros((n_chunks * FRAME_SIZE, 2))
    sig[:n_frames, 0] = left / 32768
    sig[:n_frames, 1] = right / 32768
    return sig


def save_audio_file(file_name, data: bytes):
    sig = np.frombuffer(data, dtype=BYTE_ENCODING)
    soundfile.write(file_name, sig, SAMPLING_RATE, 'PCM_16')



def cli():
    import argparse
    import asyncio
    import uvloop
    from ..common.log import parse_args_and_set_logger, logger
    from .server import start_server
    parser = argparse.ArgumentParser()
    parser.add_argument('--slave-port', type=int, default=8868, help="Port number where slave clients connect")
    parser.add_argument('--master-port', type=int, default=8878, help="Port number where the master client connects")
    parser.add_argument('--control-port', type=int, default=8858, help="Port number where the mix can be altered")
    parser.add_argument('-m', '--monitor-level', type=float,
                        help="Mix the audio file with the choir with this gain (default: don't mix it)")
    parser.add_argument('-r', '--record-audio', action="store_true", help="Keep a copy of all recordings on the server")
    parser.add_argument('-i', '--ices', help="Create a web radio by passing the given config file to ices2")

    parsed_args = parse_args_and_set_logger(parser, 'server')
    uvloop.install()
    try:
        asyncio.run(start_server(monitor_level=parsed_args.monitor_level,
                                 master_port=parsed_args.master_port,
                                 slave_port=parsed_args.slave_port,
                                 control_port=parsed_args.control_port,
                                 record_audio=parsed_args.record_audio,
                                 ices_config=parsed_args.ices))
    except KeyboardInterrupt:
        logger.info("Done.")
    exit(0)



import select
import socket
import tkinter as tk
from collections import defaultdict
from functools import partial
from hashlib import md5
from threading import Thread
from tkinter.ttk import Scale, Separator, Scrollbar, Frame
from typing import Tuple, Callable, Optional, List

from ..common.auth import AuthenticationHeader
from ..common.constants import VOCAL_GROUPS
from ..common.log import logger
from ..common.message_board import MessageBoard
from ..common.message_protocol import MessageProtocol
from ..common.pace_maker import PaceMaker
from ..common.utils import to_db, to_gain, remove_non_ascii


class DeskChannel:
    def __init__(self, top, *,
                 name: str,
                 gid: int,
                 address: Tuple,
                 key: Optional[str] = None,
                 gain: float,
                 set_gain: Callable[[Tuple, str, float], None]):
        self.name = name
        self.gid = gid
        self.key = key
        self.address = address
        self.active = True
        self._set_gain = set_gain

        self._name_label = tk.Label(top, text=name)
        self._scale = Scale(top, from_=0, to=35, length=200, command=partial(set_gain, address, key))
        self._scale.set((to_db(gain) + 25 if gain else 0))

    def grid_forget(self):
        self._name_label.grid_forget()
        self._scale.grid_forget()

    def grid_add(self, row: int):
        self._name_label.grid(row=row, column=1, sticky='W')
        self._scale.grid(row=row, column=2, sticky='W')

    def change_address(self, new_address: Tuple):
        self.address = new_address
        self._scale.config(command=partial(self._set_gain, new_address, self.key))
        self._set_gain(new_address, self.key, self._scale.get())

    def update_active(self):
        self._scale.config(state=tk.NORMAL if self.active else tk.DISABLED)


class MixingDesk:
    def __init__(self, parent, config, destroy, set_monitor_level):
        self.root = tk.Toplevel(parent)
        self.root.title("Mixing Desk")
        self.root.protocol("WM_DELETE_WINDOW", destroy)
        canvas = tk.Canvas(self.root)
        scrollbar = Scrollbar(self.root, orient="vertical", command=canvas.yview)
        scrollable_frame = Frame(canvas)
        scrollable_frame.bind(
            "<Configure>",
            lambda e: canvas.configure(
                scrollregion=canvas.bbox("all"),
                height=scrollable_frame.winfo_height(),
                width=scrollable_frame.winfo_width()
            )
        )
        canvas.create_window((0, 0), window=scrollable_frame, anchor="nw")
        canvas.configure(yscrollcommand=scrollbar.set)
        canvas.pack(side="left", fill="x", expand=True)
        scrollbar.pack(side="right", fill="y")

        self.top = scrollable_frame
        self.changes = MessageBoard()
        self.config = config

        tk.Label(self.top, text="Status:").grid(row=0, column=0, sticky='W')
        self.status = tk.Label(self.top, text="Connecting...")
        self.status.grid(row=0, column=1, sticky='W')

        Separator(self.top).grid(row=1, column=0, columnspan=3, sticky='EW', pady=10)

        tk.Label(self.top, text='Monitor').grid(row=2, column=0, sticky='W')
        tk.Label(self.top, text='(not recorded)').grid(row=2, column=1, sticky='W')
        level_scale = Scale(self.top, from_=0, to=35, length=200,
                            command=set_monitor_level)
        initial_monitor_level = float(config.get('monitor_level', '-10'))
        level_scale.set(initial_monitor_level + 35)
        level_scale.grid(row=2, column=2, sticky='EW', padx=10)

        addr = config['server'], int(config['port']) - 20  # TODO: improve
        Thread(target=self._communicate, args=addr, name='mixing desk communicator').start()

        self._channel_list: List[DeskChannel] = []
        self._separators = []

    @staticmethod
    def singer_key(name, group) -> Optional[str]:
        if not name or group not in VOCAL_GROUPS[1:]:
            return None
        return md5((name + group).encode()).hexdigest()

    def _update_desk(self, sources):
        # deactivate all
        for channel in self._channel_list:
            channel.active = False

        # match channels with same address
        by_address = {channel.address: channel for channel in self._channel_list}
        to_match = []
        for source in sources:
            address = tuple(source['id'])
            try:
                channel = by_address[address]
            except KeyError:
                to_match.append(source)
                continue
            logger.debug(f"Matched channel with {address}")
            channel.active = True

        # match channels with same key
        by_key = defaultdict(list)
        for channel in self._channel_list:
            if not channel.active and channel.key is not None:
                by_key[channel.key].append(channel)
        to_add = []
        for source in to_match:
            key = self.singer_key(source['name'], source['group'])
            if key in by_key and by_key[key]:
                logger.debug(f"Matched channel with key {key}")
                channel = by_key[key].pop()
                address = tuple(source['id'])
                channel.change_address(address)
                channel.active = True
            else:
                source['key'] = key
                to_add.append(source)

        # add_missing channels
        if to_add:
            self._delete_channels()
            for source in to_add:
                clean_name = remove_non_ascii(source['name'])
                try:
                    gid = VOCAL_GROUPS.index(source['group'])
                except ValueError:
                    gid = 0
                if source['key'] and source['key'] in self.config:  # recover saved gain
                    gain = float(self.config[source['key']])
                    logger.debug(f"found saved gain for {clean_name} ({source['key']}) in config (gain={gain})")
                else:
                    gain = source['gain']
                logger.debug(f"Adding channel for {clean_name} gid {gid} key {source['key']}")
                self._channel_list.append(DeskChannel(self.top, name=source['name'], gid=gid, key=source['key'],
                                                      address=tuple(source['id']), gain=gain, set_gain=self._set_gain))
            self._channel_list.sort(key=lambda ch: (ch.gid, ch.name))
            self._add_channels()

        # update activity
        for channel in self._channel_list:
            channel.update_active()

    def _delete_channels(self):
        for channel in self._channel_list:
            channel.grid_forget()
        for separator in self._separators:
            separator.grid_forget()
        self._separators = []

    def _add_channels(self):
        assert not self._separators
        row = 3
        gid = -1
        for channel in self._channel_list:
            if channel.gid != gid:
                gid = channel.gid
                self._separators.append(Separator(self.top))
                self._separators[-1].grid(row=row, column=0, columnspan=3, sticky='EW', pady=10)
                row += 1
                self._separators.append(tk.Label(self.top, text=VOCAL_GROUPS[gid]))
                self._separators[-1].grid(row=row, column=0, sticky='W')
            channel.grid_add(row)
            row += 1

    def _refresh_sources(self, conn):
        conn.send(MessageProtocol({"action": "list"}).to_bytes())
        logger.debug("Command sent")
        res, _, err = select.select((conn,), (), (conn,), 1.0)
        if not res:
            logger.error("failed to get list")
            raise RuntimeError("failed to get list")
        data = conn.recv(2)
        msg = MessageProtocol.from_bytes(conn.recv(MessageProtocol.get_length(data)))
        logger.debug(f"Received list with {len(msg.msg)} sources")

        self._update_desk(msg.msg)

    def _communicate(self, host, port):
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logger.debug(f"Connecting to {host}:{port}...")
        conn.connect((host, port))
        logger.debug("Connected")
        conn.send(AuthenticationHeader().to_bytes())
        self.status.config(text="Connected")
        try:
            self._refresh_sources(conn)
        except RuntimeError:
            self.status.config(text="Disconnected")
            conn.close()
            return

        # process messages
        timer = PaceMaker(0.1, len(self._channel_list))
        for source_id, gain in self.changes.messages():
            if source_id == 'update':
                try:
                    self._refresh_sources(conn)
                except RuntimeError:
                    self.status.config(text="Disconnected")
                    break
            else:
                timer.pace()
                msg = MessageProtocol({'action': 'adjust_gain', 'id': source_id, 'gain': gain})
                _, ready, err = select.select((), (conn,), (conn,), 1.0)
                if not ready:
                    logger.error("socket not ready")
                    self.status.config(text="Disconnected")
                    break
                logger.debug(f"send {msg.msg}")
                conn.send(msg.to_bytes())
        conn.close()
        logger.info("communication thread ended")

    def _set_gain(self, source_id: Tuple, key: Optional[str], fader_value):
        fader_value = float(fader_value)
        gain = to_gain(fader_value - 25) if fader_value > 0 else 0
        if key:
            self.config[key] = str(gain)
        self.changes.post_message(source_id, gain)

    def destroy(self):
        self.changes.close()
        self.root.destroy()

    def trigger_update(self):
        self.changes.post_message('update', True)



import asyncio
import numpy as np

from ..common.constants import SAMPLING_RATE, FRAME_SIZE
from ..common.log import logger

MAX_LATENCY = 120
MIN_CORRELATION = 100


def generate_test_audio():
    n = 264 * FRAME_SIZE
    freq_range = (50, 10000)
    freq = np.linspace(*freq_range, n)
    t = np.cumsum(freq) / 44100 * np.pi * 2
    sig = np.sin(t)
    return np.outer(sig, np.ones((2,)))


def calculate_latency(original, recoreded):
    c = np.correlate(recoreded, original)
    shift = np.argmax(np.abs(c))
    logger.debug(f"recorded var: {np.log10(recoreded.var())}dB, max correlation at {shift} with amplitude {c[shift]}")
    if abs(c[shift]) < 100:
        return None
    return shift / FRAME_SIZE


async def main(sig):
    import sounddevice
    assert len(sig) % FRAME_SIZE == 0
    to_play = np.concatenate((sig, np.zeros((MAX_LATENCY * FRAME_SIZE, 2))))
    res = np.empty((len(to_play),), dtype=np.float32)
    loop = asyncio.get_event_loop()
    end = asyncio.Event()

    class AudioIndex:
        idx: int = 0

    cur = AudioIndex()

    def callback(indata: np.ndarray, outdata: np.ndarray, frames: int, _, status: sounddevice.CallbackFlags):
        assert frames == FRAME_SIZE
        if status.output_underflow:
            print("Underflow")
            raise sounddevice.CallbackAbort
        assert status.input_underflow or not status
        res[cur.idx:cur.idx + len(indata)] = indata[:, 0]
        outdata[:, :] = to_play[cur.idx:cur.idx + FRAME_SIZE, :]
        cur.idx += FRAME_SIZE
        if cur.idx >= len(to_play):
            raise sounddevice.CallbackStop

    def finished_callback():
        loop.call_soon_threadsafe(end.set)

    stream = sounddevice.Stream(samplerate=SAMPLING_RATE,
                                blocksize=FRAME_SIZE,
                                channels=(1, 2),
                                dtype="float32",
                                callback=callback,
                                finished_callback=finished_callback)

    stream.start()
    await end.wait()
    stream.stop()
    stream.close()

    latency = calculate_latency(sig.mean(axis=1), res)

    print(f"Measured latency: {latency}")


def cli():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', default='assets/sound/sine.wav')
    parser.add_argument('-s', '--sine', action='store_true')
    parsed_args = parser.parse_args()

    if parsed_args.sine:
        sig = generate_test_audio()
    else:
        from ..common.audio_files import load_audio_file
        sig = load_audio_file(parsed_args.input).mean(axis=1)
    asyncio.run(main(sig))

    parser.exit(0)



import numpy as np
from enum import IntEnum

SAMPLING_RATE = 44100
FRAME_SIZE = 512
BYTE_ENCODING = np.dtype('i2')
SATURATION_LEVEL = 0.999
CHUNKS_PER_SECOND = SAMPLING_RATE / FRAME_SIZE
CHUNKS_IN_FLIGHT = int(6 * CHUNKS_PER_SECOND)
SLAVE_PLAYBACK_BUFFER_SIZE = int(CHUNKS_PER_SECOND)
MASTER_PLAYBACK_BUFFER_SIZE = int(2 * CHUNKS_PER_SECOND)

KEEPALIVE_INTERVAL = 10
CONNECTION_TIMEOUT = 30


# TODO: replace idx = 0, -1 and -2 everywhere.
class FrameType(IntEnum):
    EOF = -1  # for lead track
    CHECK_IN = -1  # for slave client stream
    START = 0
    ABORT = -2
    KEEPALIVE = -3


VOCAL_GROUPS = ['No group', 'Soprano', 'Mezzo', 'Alto', 'Tenor', 'Baritone', 'Bass']
